import os
import zipfile
from os.path import join

import pandas as pd
from pyproj import Transformer

from aequilibrae.project.database_connection import database_connection
from aequilibrae.transit.functions.get_srid import get_srid
from aequilibrae.transit.gtfs_writer import write_routes, write_agencies, write_fares
from aequilibrae.transit.gtfs_writer import write_stops, write_trips, write_stop_times, write_shapes
from aequilibrae.transit.route_system_reader import read_agencies, read_patterns
from aequilibrae.transit.route_system_reader import read_stop_times, read_stops, read_trips, read_routes
from aequilibrae.utils.db_utils import commit_and_close


class RouteSystem:
    def __init__(self, database_path):
        self.__database_path = database_path

        self.agencies = []
        self.stops = []
        self.routes = []
        self.trips = []
        self.patterns = []
        self.stop_times = pd.DataFrame([])

        self.fare_attributes = []
        self.fare_rules = []
        self.zones = []

        self.transformer = Transformer.from_crs(f"epsg:{get_srid()}", "epsg:4326", always_xy=True)

    def load_route_system(self):
        with commit_and_close(database_connection(join(self.__database_path, "public_transport.sqlite"))) as conn:
            self._read_agencies(conn)
            self._read_stops(conn)
            self._read_routes(conn)
            self._read_patterns(conn)
            self._read_trips(conn)
            self._read_stop_times(conn)

    def _read_agencies(self, conn):
        self.agencies = read_agencies(conn)

    def _read_stops(self, conn):
        self.stops = read_stops(conn, self.transformer)

    def _read_routes(self, conn):
        self.routes = read_routes(conn)

    def _read_patterns(self, conn):
        self.patterns = self.patterns or read_patterns(conn, self.transformer)

    def _read_trips(self, conn):
        self.trips = self.trips or read_trips(conn)

    def _read_stop_times(self, conn):
        self.stop_times = read_stop_times(conn)

    def write_GTFS(self, path_to_folder: str):
        """ """

        with commit_and_close(database_connection(join(self.__database_path, "public_transport.sqlite"))) as conn:
            write_agencies(self.agencies, path_to_folder)
            write_stops(self.stops, path_to_folder)
            write_routes(self.routes, path_to_folder)
            write_shapes(self.patterns, path_to_folder)

            write_trips(self.trips, path_to_folder, conn)
            write_stop_times(self.stop_times, path_to_folder)
            write_fares(path_to_folder, conn)
            self._zip_feed(path_to_folder)

    def _zip_feed(self, path_to_folder: str):
        filename = join(path_to_folder, "polaris_gtfs.zip")
        files = [
            "agency",
            "stops",
            "routes",
            "trips",
            "stop_times",
            "calendar",
            "shapes",
            "fare_attributes",
            "fare_rules",
        ]
        with zipfile.ZipFile(filename, mode="w", compression=zipfile.ZIP_DEFLATED) as zip_file:
            for file in files:
                zip_file.write(join(path_to_folder, f"{file}.txt"), f"{file}.txt")
                os.unlink(join(path_to_folder, f"{file}.txt"))



"""
AequilibraE run module.

This module is dynamically imported when the `project.run` property is accessed. Objects named within `parameters.yml`
under the `run` heading will have their arguments partially applied via `functools.partial` and will replace the objects
within this module.

Not all objects within this module must be named `parameters.yml`. If an object is named within `parameters.yml`, then it
must exist within this module otherwise a `RuntimeError` will be raised.

Functions should use the `get_active_project()` function to obtain a reference to the current project.

State within this module should be avoided as this file may be run multiple times.
"""

import numpy as np
import pandas as pd

from aequilibrae.context import get_active_project


def matrix_summary():
    """
    Compute summary statistics about the matrices registered with the active project.

    If no matrices are registered an empty dictionary will be returned.
    """
    project = get_active_project()
    mats = project.matrices
    df = mats.list()

    res = {}
    for name in df["name"]:
        mat = mats.get_matrix(name)
        mat.computational_view()

        stats = {}
        for i, nm in enumerate(mat.view_names):
            array = mat.matrix_view[:, :] if len(mat.view_names) == 1 else mat.matrix_view[:, :, i]
            stats[nm] = {
                "total": np.sum(array),
                "min": np.min(array),
                "max": np.max(array),
                "nnz": (array != 0.0).sum(),
            }
        res[name] = stats

    return res


def graph_summary():
    """
    Compute summary statistics about the graphs built within the active project.

    If no graphs have been built an empty dictionary will be returned.
    """
    project = get_active_project()
    graphs = project.network.graphs

    return {
        k: {
            "num_links": v.num_links,
            "num_nodes": v.num_nodes,
            "num_zones": v.num_zones,
            "compact_num_links": v.compact_num_links,
            "compact_num_nodes": v.compact_num_nodes,
        }
        for k, v in graphs.items()
    }


def results_summary():
    """
    Read the results table from the project database.
    """
    project = get_active_project()

    sql = """SELECT * from results;"""
    with project.db_connection as conn:
        return pd.read_sql(sql, conn)


def example_function_with_kwargs(arg1: str = None, **kwargs):
    """
    An example function to demonstrate the argument application via parameters.yml.
    """
    if arg1 is None:
        arg1 = "default argument"

    print("arg1:", arg1)
    print("kwargs:", kwargs)



import sqlite3

import pandas as pd
from aequilibrae.utils.get_table import get_table

# from polarislib.network.data import DataTableStorage
from aequilibrae.transit.transit_elements import Trip


def read_trips(conn: sqlite3.Connection):
    data = get_table("transit_trips", conn).reset_index()
    data.drop(columns=["pce", "seated_capacity", "design_capacity", "total_capacity", "is_artic"], inplace=True)
    data.drop(columns=["number_of_cars"], inplace=True)

    pats = pd.read_sql("Select pattern_id, route_id from Transit_Patterns", conn)
    data = data.merge(pats, on="pattern_id")
    data.trip = data.trip.astype(str)
    data.rename(
        columns={
            "trip": "trip_headsign",
            "dir": "direction_id",
            "pattern_id": "shape_id",
        },
        inplace=True,
    )
    data = data.assign(service_id=data.shape_id)
    return [Trip().from_row(dt) for _, dt in data.iterrows()]



from aequilibrae.utils.db_utils import read_and_close


def find_table_fields(table_name, conn=None, db_path=None):
    with conn or read_and_close(db_path, spatial=True) as conn:
        structure = conn.execute(f"pragma table_info({table_name})").fetchall()
    geotypes = ["LINESTRING", "POINT", "POLYGON", "MULTIPOLYGON"]
    fields = [x[1].lower() for x in structure]
    geotype = geo_field = None
    for x in structure:
        if x[2].upper() in geotypes:
            geotype = x[2]
            geo_field = x[1]
            break
    if geo_field is not None:
        fields = [x for x in fields if x != geo_field.lower()]

    return fields, geotype, geo_field



import logging
import re
from os.path import join, dirname, realpath
from sqlite3 import Connection
from aequilibrae.project.database_connection import database_connection
from aequilibrae.utils.db_utils import commit_and_close

req_link_flds = ["link_id", "a_node", "b_node", "direction", "distance", "modes", "link_type"]
req_node_flds = ["node_id", "is_centroid"]
protected_fields = ["ogc_fid", "geometry"]


def initialize_tables(project, db_type: str) -> None:
    with commit_and_close(database_connection(db_type)) as conn:
        create_base_tables(conn, project.logger, db_type)
        add_triggers(conn, project.logger, db_type)


def create_base_tables(conn: Connection, logger: logging.Logger, db_type: str) -> None:
    spec_folder = join(dirname(realpath(__file__)), "database_specification", db_type, "tables")
    with open(join(spec_folder, "table_list.txt"), "r") as file_list:
        all_tables = file_list.readlines()
    all_tables = [x.rstrip() for x in all_tables]
    for f in all_tables:
        qry_file = join(spec_folder, f"{f}.sql")
        run_queries_from_sql_file(conn, logger, qry_file)


def add_triggers(conn: Connection, logger: logging.Logger, db_type: str) -> None:
    """Adds consistency triggers to the project"""
    spec_folder = join(dirname(realpath(__file__)), "database_specification", db_type, "triggers")
    with open(join(spec_folder, "triggers_list.txt"), "r") as file_list:
        all_trigger_sets = file_list.readlines()
    all_trigger_sets = [x.rstrip() for x in all_trigger_sets]
    for f in all_trigger_sets:
        qry_file = join(spec_folder, f"{f}.sql")
        run_queries_from_sql_file(conn, logger, qry_file)


def remove_triggers(conn: Connection, logger: logging.Logger, db_type: str) -> None:
    spec_folder = join(dirname(realpath(__file__)), "database_specification", db_type, "triggers")
    with open(join(spec_folder, "triggers_list.txt"), "r") as file_list:
        all_trigger_sets = file_list.readlines()

    create_drop_regex = re.compile(r"create\s+trigger\s+(\w+)", flags=re.I)
    for table in all_trigger_sets:
        qry_file = join(spec_folder, f"{table.rstrip()}.sql")

        with open(qry_file, "r") as sql_file:
            query_list = sql_file.read()

        # Running one query/command at a time helps debugging in the case a particular command fails
        for cmd in query_list.split("--#"):
            for qry in cmd.split("\n"):
                if qry[:2] == "--":
                    continue
                while "  " in qry:
                    qry = qry.replace("  ", " ")

                m = re.search(create_drop_regex, qry)
                if m:
                    try:
                        conn.execute(f"drop trigger if exists {m.group(1).lower()}")
                    except Exception as e:
                        logger.error(f"Failed removing triggers table - > {e.args}")
                        logger.error(f"Point of failure - > {qry}")


def run_queries_from_sql_file(conn: Connection, logger: logging.Logger, qry_file: str) -> None:
    with open(qry_file, "r") as sql_file:
        query_list = sql_file.read()

    # Running one query/command at a time helps debugging in the case a particular command fails
    for cmd in query_list.split("--#"):
        try:
            conn.execute(cmd)
        except Exception as e:
            msg = f"Error running SQL command: {e.args}"
            logger.error(msg)
            logger.info(cmd)
            raise e



import sqlite3
from pathlib import Path

from aequilibrae.context import get_active_project
from aequilibrae.utils.spatialite_utils import connect_spatialite


def database_connection(db_type: str, project_path=None) -> sqlite3.Connection:
    return connect_spatialite(database_path(db_type, project_path))


def database_path(db_type: str, project_path=None) -> Path:
    project_path = project_path or get_active_project().project_base_path
    db = "public_transport" if db_type == "transit" else "project_database"
    return Path(project_path) / f"{db}.sqlite"



from libcpp.algorithm cimport min_element, sort, lower_bound
from libc.math cimport INFINITY, exp, pow, log
from libcpp cimport nullptr
from libcpp.memory cimport shared_ptr, make_shared

from cython.operator cimport dereference as d
from cython.operator cimport postincrement as inc

import pandas as pd
import numpy as np

import cython

from aequilibrae.log import logger


@cython.embedsignature(True)
cdef class RouteChoiceSetResults:
    """
    This class is supposed to help manage and compute the results of the route choice set generation. It also
    provides method to perform an assignment and link loading.
    """

    def __init__(
            self,
            demand: GeneralisedCOODemand,
            cutoff_prob: float,
            beta: float,
            num_links: int,
            double[:] cost_view,
            unsigned int [:] mapping_idx,
            int64_t [::] mapping_data,
            store_results: bool = True,
            perform_assignment: bool = True,
    ):
        """
        :Arguments:
            **demand** (`obj`: GeneralisedCOODemand): A GeneralisedCOODemand object stores the ODs pairs and various
              demand values in a COO form. No verification of these is performed here.

            **cutoff_prob** (`obj`: float): The cut-off probability for the inverse binary logit filter.

            **beta** (`obj`: float): The beta parameter for the path-sized logit.

            **store_results** (`obj`: bool): Whether or not to store the route set computation results. At a minimum
              stores the route sets per OD. If `perform_assignment` is True then the assignment results are stored as
              well.

            **perform_assignment** (`obj`: bool): Whether or not to perform a path-sized logit assignment.

        NOTE: This class makes no attempt to be thread safe when improperly accessed. Multithreaded accesses should be
        coordinated to not collide. Each index of `ods` should only ever be accessed by a single thread.

        NOTE: Depending on `store_results` the behaviour of accessing a single `ods` index multiple times will
        differ. When True the previous internal buffers will be reused. This will highly likely result incorrect
        results. When False some new internal buffers will used, link loading results will still be incorrect. Thus A
        SINGLE `ods` INDEX SHOULD NOT BE ACCESSED MULTIPLE TIMES.
        """

        if not store_results and not perform_assignment:
            raise ValueError("either `store_results` or `perform_assignment` must be True")

        self.demand = demand
        self.cutoff_prob = cutoff_prob
        self.beta = beta
        self.store_results = store_results
        self.perform_assignment = perform_assignment
        self.cost_view = cost_view
        self.mapping_idx = mapping_idx
        self.mapping_data = mapping_data
        self.table = None

        cdef size_t size = self.demand.ods.size()

        # As the objects are attribute of the extension class they will be allocated before the object is
        # initialised. This ensures that accessing them is always valid and that they are just empty. We resize the ones
        # we will be using here and allocate the objects they store for the same reasons.
        #
        # We can't know how big they will be so we'll need to resize later as well.
        if self.store_results:
            self.__route_vecs.resize(size)
            for i in range(size):
                self.__route_vecs[i] = make_shared[RouteVec_t]()

        if self.perform_assignment and self.store_results:
            self.__cost_set.resize(size)
            self.__mask_set.resize(size)
            self.__path_overlap_set.resize(size)
            self.__prob_set.resize(size)
            for i in range(size):
                self.__cost_set[i] = make_shared[vector[double]]()
                self.__mask_set[i] = make_shared[vector[bint]]()
                self.__path_overlap_set[i] = make_shared[vector[double]]()
                self.__prob_set[i] = make_shared[vector[double]]()

    def write(self, where):
        table = self.make_df_from_results()

        engine = pd.io.parquet.get_engine('auto').__class__
        if (engine.__module__, engine.__name__) == ("pandas.io.parquet", "PyArrowImpl"):
            kwargs = dict(
                # can't provide partitioning_flavor and partition_cols through the Pandas API
                use_threads=True,
                existing_data_behavior="overwrite_or_ignore",
                file_visitor=lambda written_file: logger.info(f"Wrote partition dataset at {written_file.path}")
            )
        elif (engine.__module__, engine.__name__) == ("pandas.io.parquet", "FastParquetImpl"):
            logger.info("FastParquet back-end doesn't support individual partition logging, writing table now...")
            kwargs = dict(
                file_scheme="hive",
                # no threads option
                append=False,
                # no visitor option
            )
            logger.warn(
                "FastParquet back-end doesn't support writing a NumPy arrays as Parquet list types, converting to Python lists. "
                "Watch out for memory consumption..."
            )
            table["route set"] = table["route set"].map(lambda x: x.tolist())
        else:
            raise RuntimeError(
                "encountered unknown Pandas parquet engine, please report this as a bug to the AequilibraE issues page"
            )

        table.to_parquet(
            path=where,
            compression="zstd",
            index=False,
            partition_cols=["origin id"],
            **kwargs,
        )

    @classmethod
    def read_dataset(cls, where):
        df = pd.read_parquet(where, partitioning="hive")
        df["origin id"] = df["origin id"].astype(df["destination id"].dtype)

        # FastParquet is stupid and encodes Parquet list objects as json strings!!!
        is_json_encoded = df["route set"].map(lambda x: isinstance(x, (str, bytes)))
        if is_json_encoded.any():
            logger.warn("Found JSON encoded route sets. Parsing into a NumPy array...")
            if not is_json_encoded.all():
                raise TypeError(
                    f"route sets must either be encoded properly as list[int64], or json lists (by FastParquet). The two cannot be mixed"
                )

            import json
            df["route set"] = df["route set"].map(lambda x: np.array(json.loads(x), dtype="int64"))

        return df

    @staticmethod
    cdef void route_set_to_route_vec(RouteVec_t &route_vec, RouteSet_t &route_set) noexcept nogil:
        """
        Transform a set of raw pointers to routes (vectors) into a vector of unique points to
        routes.
        """
        cdef vector[long long] *route

        route_vec.reserve(route_set.size())
        for route in route_set:
            route_vec.emplace_back(route)

        # We now drop all references to those raw pointers. The unique pointers now own those vectors.
        route_set.clear()

    cdef shared_ptr[RouteVec_t] get_route_vec(RouteChoiceSetResults self, size_t i) noexcept nogil:
        """
        Return either a new empty RouteSet_t, or the RouteSet_t (initially empty) corresponding to a OD pair index.

        If `self.store_results` is False no attempt is made to store the route set. The caller is responsible for
        maintaining a reference to it.

        Requires that 0 <= i < self.ods.size().
        """
        if self.store_results:
            # All elements of self.__route_vecs have been initialised in self.__init__.
            return self.__route_vecs[i]
        else:
            # We make a new empty RouteSet_t here, we don't attempt to store it.
            return make_shared[RouteVec_t]()

    cdef shared_ptr[vector[double]] __get_cost_set(RouteChoiceSetResults self, size_t i) noexcept nogil:
        return self.__cost_set[i] if self.store_results else make_shared[vector[double]]()

    cdef shared_ptr[vector[bint]] __get_mask_set(RouteChoiceSetResults self, size_t i) noexcept nogil:
        return self.__mask_set[i] if self.store_results else make_shared[vector[bint]]()

    cdef shared_ptr[vector[double]] __get_path_overlap_set(RouteChoiceSetResults self, size_t i) noexcept nogil:
        return self.__path_overlap_set[i] if self.store_results else make_shared[vector[double]]()

    cdef shared_ptr[vector[double]] get_prob_vec(RouteChoiceSetResults self, size_t i) noexcept nogil:
        return self.__prob_set[i] if self.store_results else make_shared[vector[double]]()

    cdef shared_ptr[vector[double]] compute_result(
        RouteChoiceSetResults self,
        size_t i,
        RouteVec_t &route_set,
        bint *found_zero_cost,
        size_t thread_id
    ) noexcept nogil:
        """
        Compute the desired results for the OD pair index with the provided route set. The route set is required as
        an argument here to facilitate not storing them. The route set should correspond to the provided OD pair index,
        however that is not enforced.

        Requires that 0 <= i < self.ods.size().

        Returns a shared pointer to the probability vector.
        """
        cdef:
            shared_ptr[vector[double]] cost_vec
            shared_ptr[vector[bint]] route_mask
            vector[long long] keys, counts
            shared_ptr[vector[double]] path_overlap_vec
            shared_ptr[vector[double]] prob_vec

        if not self.perform_assignment:
            # If we're not performing an assignment then we must be storing the routes and the routes most already be
            # stored when they were acquired, thus we don't need to do anything here.
            return make_shared[vector[double]]()

        cost_vec = self.__get_cost_set(i)
        route_mask = self.__get_mask_set(i)
        path_overlap_vec = self.__get_path_overlap_set(i)
        prob_vec = self.get_prob_vec(i)

        self.compute_cost(d(cost_vec), route_set, self.cost_view, found_zero_cost)
        self.compute_mask(d(route_mask), d(cost_vec))
        self.compute_frequency(keys, counts, route_set, d(route_mask))
        self.compute_path_overlap(
            d(path_overlap_vec),
            route_set,
            keys,
            counts,
            d(cost_vec),
            d(route_mask),
            self.cost_view
        )
        self.compute_prob(d(prob_vec), d(cost_vec), d(path_overlap_vec), d(route_mask))

        return prob_vec

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void compute_cost(
        RouteChoiceSetResults self,
        vector[double] &cost_vec,
        const RouteVec_t &route_set,
        const double[:]
        cost_view,
        bint *found_zero_cost
    ) noexcept nogil:
        """Compute the cost each route."""
        cdef:
            # Scratch objects
            double cost
            long long link
            size_t i

        cost_vec.resize(route_set.size())

        found_zero_cost[0] = False
        for i in range(route_set.size()):
            cost = 0.0
            for link in d(route_set[i]):
                cost = cost + cost_view[link]

            cost_vec[i] = cost
            if cost == 0.0:
                found_zero_cost[0] = True

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void compute_mask(
        RouteChoiceSetResults self,
        vector[bint] &route_mask,
        const vector[double] &total_cost
    ) noexcept nogil:
        """
        Computes a binary logit between the minimum cost path and each path, if the total cost is greater than the
        minimum + the difference in utilities required to produce the cut-off probability then the route is excluded
        from the route set.
        """
        cdef:
            bint found_zero_cost = False
            size_t i

            vector[double].const_iterator min = min_element(total_cost.cbegin(), total_cost.cend())
            double cutoff_cost

        if min == total_cost.cend():
            cutoff_cost = INFINITY
        else:
            cutoff_cost = d(min) + inverse_binary_logit(self.cutoff_prob, 0.0, 1.0)

        route_mask.resize(total_cost.size())

        # The route mask should be True for the routes we wish to include.
        for i in range(total_cost.size()):
            if total_cost[i] == 0.0:
                found_zero_cost = True
                break
            elif total_cost[i] <= cutoff_cost:
                route_mask[i] = True

        if found_zero_cost:
            # If we've found a zero cost path we must abandon the whole route set.
            for i in range(total_cost.size()):
                route_mask[i] = False
        elif min != total_cost.cend():
            # Always include the min element. It should already be but I don't trust floating math to do this correctly.
            # But only if there actually was a min element (i.e. empty route set)
            route_mask[min - total_cost.cbegin()] = True

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void compute_frequency(
        RouteChoiceSetResults self,
        vector[long long] &keys,
        vector[long long] &counts,
        const RouteVec_t &route_set,
        const vector[bint] &route_mask
    ) noexcept nogil:
        """
        Compute a frequency map for each route with the route_mask applied.

        Each node at index i in the first returned vector has frequency at index i in the second vector.
        """
        cdef:
            vector[long long] link_union
            vector[long long].const_iterator union_iter

            # Scratch objects
            size_t length, count, i
            long long link

        # When calculating the frequency of routes, we need to exclude those not in the mask.
        length = 0
        for i in range(route_set.size()):
            # We do so here ...
            if not route_mask[i]:
                continue

            length = length + d(route_set[i]).size()
        link_union.reserve(length)

        for i in range(route_set.size()):
            # ... and here.
            if not route_mask[i]:
                continue

            link_union.insert(link_union.end(), d(route_set[i]).begin(), d(route_set[i]).end())

        sort(link_union.begin(), link_union.end())

        union_iter = link_union.cbegin()
        while union_iter != link_union.cend():
            count = 0
            link = d(union_iter)
            while union_iter != link_union.cend() and link == d(union_iter):
                count = count + 1
                inc(union_iter)

            keys.push_back(link)
            counts.push_back(count)

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    @cython.cdivision(True)
    cdef void compute_path_overlap(
        RouteChoiceSetResults self,
        vector[double] &path_overlap_vec,
        const RouteVec_t &route_set,
        const vector[long long] &keys,
        const vector[long long] &counts,
        const vector[double] &total_cost,
        const vector[bint] &route_mask,
        const double[:] cost_view
    ) noexcept nogil:
        """
        Compute the path overlap figure based on the route cost and frequency.

        Notation changes:
            a: link
            t_a: cost_view
            c_i: total_costs
            A_i: route
            sum_{k in R}: delta_{a,k}: freq_set
        """
        cdef:
            # Scratch objects
            vector[long long].const_iterator link_iter
            double path_overlap
            long long link
            size_t i

        path_overlap_vec.resize(route_set.size())

        for i in range(route_set.size()):
            # Skip masked routes
            if not route_mask[i]:
                continue

            path_overlap = 0.0
            for link in d(route_set[i]):
                # We know the frequency table is ordered and contains every link in the union of the routes.
                # We want to find the index of the link, and use that to look up it's frequency
                link_iter = lower_bound(keys.cbegin(), keys.cend(), link)

                # lower_bound returns keys.end() when no link is found.
                # This /should/ never happen.
                if link_iter == keys.cend():
                    continue
                path_overlap = path_overlap + cost_view[link] / counts[link_iter - keys.cbegin()]

            path_overlap_vec[i] = path_overlap / total_cost[i]

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    @cython.cdivision(True)
    cdef void compute_prob(
        RouteChoiceSetResults self,
        vector[double] &prob_vec,
        const vector[double] &total_cost,
        const vector[double] &path_overlap_vec,
        const vector[bint] &route_mask
    ) noexcept nogil:
        """Compute a probability for each route in the route set based on the path overlap."""
        cdef:
            # Scratch objects
            double inv_prob
            size_t i, j

        prob_vec.resize(total_cost.size())

        # Beware when refactoring the below, the scale of the costs may cause floating point errors. Large costs will
        # lead to NaN results
        for i in range(total_cost.size()):
            # The probability of choosing a route that has been masked out is 0.
            if not route_mask[i]:
                continue

            inv_prob = 0.0
            for j in range(total_cost.size()):
                # We must skip any other routes that are not included in the mask otherwise our probabilities won't
                # add up.
                if not route_mask[j]:
                    continue

                inv_prob = inv_prob + pow(path_overlap_vec[j] / path_overlap_vec[i], self.beta) \
                    * exp((total_cost[i] - total_cost[j]))  # Assuming theta=1.0

            prob_vec[i] = 1.0 / inv_prob

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef object make_df_from_results(RouteChoiceSetResults self):
        """
        Construct an pd.DataFrame from the C++ stdlib structures.

        Compressed link IDs are expanded to full network link IDs.
        """

        if self.table is not None:
            return self.table
        elif not self.store_results:
            raise RuntimeError("route set table construction requires `store_results` is True")

        cdef:
            size_t link, tmp, n_routes
            bint have_assignment_results = self.perform_assignment and self.store_results

        columns = {
            "origin id": [],
            "destination id": [],
        }
        route_set_col = []  # We treat this one differently when constructing it

        if have_assignment_results:
            columns["cost"] = []
            columns["mask"] = []
            columns["path overlap"] = []
            columns["probability"] = []

        if have_assignment_results:
            for i in range(self.demand.ods.size()):
                n_routes = d(self.__route_vecs[i]).size()
                if not d(self.__route_vecs[i]).size():  # If there's no routes to add just skip these.
                    continue

                # When assigning from df, the cost, mask, and path overlap vectors may be empty
                tmp = d(self.__cost_set[i]).size()
                columns["cost"].append(
                    np.asarray(<double[:tmp]>d(self.__cost_set[i]).data()) if tmp else np.zeros(n_routes, dtype="float64")
                )

                tmp = d(self.__mask_set[i]).size()
                columns["mask"].append(
                    np.asarray(<bint[:tmp]>d(self.__mask_set[i]).data()) if tmp else np.ones(n_routes, dtype="bool")
                )

                tmp = d(self.__path_overlap_set[i]).size()
                columns["path overlap"].append(
                    np.asarray(<double[:tmp]>d(self.__path_overlap_set[i]).data()) if tmp else np.zeros(n_routes, dtype="float64")
                )

                tmp = d(self.__prob_set[i]).size()
                columns["probability"].append(
                    np.asarray(<double[:tmp]>d(self.__prob_set[i]).data())
                )

        for i in range(self.demand.ods.size()):
            route_set = self.__route_vecs[i]

            columns["origin id"].append(np.full(d(route_set).size(), self.demand.ods[i].first, "uint32"))
            columns["destination id"].append(np.full(d(route_set).size(), self.demand.ods[i].second, "uint32"))

            # Instead of constructing a "list of lists" style object for storing the route sets we instead will
            # construct one big array of link IDs with a corresponding offsets array that indicates where each new row
            # (path) starts.
            for j in range(d(route_set).size()):

                links = []
                for link in d(d(route_set)[j]):
                    # Translate the compressed link IDs in route to network link IDs, this is a 1:n mapping
                    links.append(np.asarray(self.mapping_data[self.mapping_idx[link]:self.mapping_idx[link + 1]]))

                route_set_col.append(np.hstack(links))

        columns = {k: np.hstack(v, casting="no") if len(v) else np.array([], dtype=v.dtype) for k, v in columns.items()}
        columns["route set"] = route_set_col

        self.table = pd.DataFrame(columns)
        return self.table


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.initializedcheck(False)
@cython.cdivision(True)
cdef double inverse_binary_logit(double prob, double beta0, double beta1) noexcept nogil:
    if prob == 1.0:
        return INFINITY
    elif prob == 0.0:
        return -INFINITY
    else:
        return (log(prob / (1.0 - prob)) - beta0) / beta1



from cython.parallel cimport prange


def sum_axis1(totals, multiples, cores):
    cdef int c = cores
    cdef double [:] totals_view = totals
    cdef double [:, :] multiples_view = multiples

    with nogil:
        sum_axis1_cython(totals_view, multiples_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void sum_axis1_cython(double[:] totals,
                            double[:, :] multiples,
                            int cores) noexcept nogil:
    cdef long long i, j
    cdef long long l = totals.shape[0]
    cdef long long k = multiples.shape[1]

    # TODO: Use prange with use_threads_if when Cython 3.1 is released
    for i in range(l):
        totals[i] = 0
        for j in range(k):
            totals[i] += multiples[i, j]


def sum_a_times_b_minus_c(array1, array2, array3, cores):
    cdef int c = cores
    cdef double result
    cdef double [:] array1_view = array1
    cdef double [:] array2_view = array2
    cdef double [:] array3_view = array3

    with nogil:
        result = sum_a_times_b_minus_c_cython(array1_view, array2_view, array3_view, c)
    return result


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef double sum_a_times_b_minus_c_cython(
    double[:] array1,
    double[:] array2,
    double[:] array3,
    int cores
) noexcept nogil:
    cdef long long i
    cdef double row_result
    cdef double result = 0.0
    cdef long long l = array1.shape[0]

    # TODO: Use prange with use_threads_if when Cython 3.1 is released
    for i in range(l):
        row_result = array1[i] * (array2[i] - array3[i])
        result += row_result

    return result


def linear_combination_1d(results, array1, array2, stepsize, cores):
    cdef double stpsz
    cdef int c = cores

    stpsz = float(stepsize)
    cdef double [:] results_view = results
    cdef double [:] array1_view = array1
    cdef double [:] array2_view = array2

    with nogil:
        linear_combination_cython_1d(stpsz, results_view, array1_view, array2_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void linear_combination_cython_1d(
    double stepsize,
    double[:] results,
    double[:] array1,
    double[:] array2,
    int cores
) noexcept nogil:
    cdef long long i
    cdef long long l = results.shape[0]

    # TODO: Use prange with use_threads_if when Cython 3.1 is released
    for i in range(l):
        results[i] = array1[i] * stepsize + array2[i] * (1.0 - stepsize)


def linear_combination(results, array1, array2, stepsize, cores):
    cdef double stpsz
    cdef int c = cores

    stpsz = float(stepsize)
    cdef double [:, :] results_view = results
    cdef double [:, :] array1_view = array1
    cdef double [:, :] array2_view = array2

    with nogil:
        linear_combination_cython(stpsz, results_view, array1_view, array2_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void linear_combination_cython(
    double stepsize,
    double[:, :] results,
    double[:, :] array1,
    double[:, :] array2,
    int cores
) noexcept nogil:
    cdef long long i, j
    cdef long long l = results.shape[0]
    cdef long long k = results.shape[1]

    # TODO: Use prange with use_threads_if when Cython 3.1 is released
    for i in range(l):
        for j in range(k):
            results[i, j] = array1[i, j] * stepsize + array2[i, j] * (1.0 - stepsize)


def linear_combination_skims(results, array1, array2, stepsize, cores):
    cdef double stpsz
    cdef int c = cores

    stpsz = float(stepsize)
    cdef double [:, :, :] results_view = results
    cdef double [:, :, :] array1_view = array1
    cdef double [:, :, :] array2_view = array2

    linear_combination_skims_cython(stpsz, results_view, array1_view, array2_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void linear_combination_skims_cython(
    double stepsize,
    double[:, :, :] results,
    double[:, :, :] array1,
    double[:, :, :] array2,
    int cores
) noexcept:
    cdef long long i, j, k
    cdef long long a = results.shape[0]
    cdef long long b = results.shape[1]
    cdef long long c = results.shape[2]

    for i in prange(a, nogil=True, num_threads=cores):
        for k in range(c):
            for j in range(b):
                results[i, j, k] = array1[i, j, k] * stepsize + array2[i, j, k] * (1.0 - stepsize)


def triple_linear_combination(results, array1, array2, array3, stepsizes, cores):
    cdef int c = cores

    cdef double [:, :] results_view = results
    cdef double [:, :] array1_view = array1
    cdef double [:, :] array2_view = array2
    cdef double [:, :] array3_view = array3
    cdef double [:] stpsz_view = stepsizes

    triple_linear_combination_cython(stpsz_view, results_view, array1_view, array2_view, array3_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void triple_linear_combination_cython(
    double [:] stepsizes,
    double[:, :] results,
    double[:, :] array1,
    double[:, :] array2,
    double[:, :] array3,
    int cores
) noexcept:
    cdef long long i, j
    cdef long long l = results.shape[0]
    cdef long long k = results.shape[1]

    for i in prange(l, nogil=True, num_threads=cores):
        for j in range(k):
            results[i, j] = array1[i, j] * stepsizes[0] + array2[i, j] * stepsizes[1] + array3[i, j] * stepsizes[2]


def triple_linear_combination_skims(results, array1, array2, array3, stepsizes, cores):
    cdef int c = cores

    cdef double [:, :, :] results_view = results
    cdef double [:, :, :] array1_view = array1
    cdef double [:, :, :] array2_view = array2
    cdef double [:, :, :] array3_view = array3
    cdef double [:] stpsz_view = stepsizes

    triple_linear_combination_cython_skims(stpsz_view, results_view, array1_view, array2_view, array3_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void triple_linear_combination_cython_skims(
    double [:] stepsizes,
    double[:, :, :] results,
    double[:, :, :] array1,
    double[:, :, :] array2,
    double[:, :, :] array3,
    int cores
) noexcept:
    cdef long long i, j, k
    cdef long long a = results.shape[0]
    cdef long long b = results.shape[1]
    cdef long long c = results.shape[2]

    for i in prange(a, nogil=True, num_threads=cores):
        for k in range(c):
            for j in range(b):
                results[i, j, k] = array1[i, j, k] * stepsizes[0] + array2[i, j, k] * stepsizes[1]  + \
                                   array3[i, j, k] * stepsizes[2]


def copy_one_dimension(target, source, cores):
    cdef int c = cores

    cdef double [:] target_view = target
    cdef double [:] source_view = source

    copy_one_dimension_cython(target_view, source_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void copy_one_dimension_cython(double[:] target, double[:] source, int cores) noexcept:
    cdef long long i
    cdef long long l = target.shape[0]

    for i in prange(l, nogil=True, num_threads=cores):
        target[i] = source[i]


def copy_two_dimensions(target, source, cores):
    cdef int c = cores

    cdef double [:, :] target_view = target
    cdef double [:, :] source_view = source

    copy_two_dimensions_cython(target_view, source_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void copy_two_dimensions_cython(double[:, :] target, double[:, :] source, int cores) noexcept:
    cdef long long i, j
    cdef long long l = target.shape[0]
    cdef long long k = target.shape[1]

    for i in prange(l, nogil=True, num_threads=cores):
        for j in range(k):
            target[i, j] = source[i, j]


def copy_three_dimensions(target, source, cores):
    cdef int c = cores

    cdef double [:, :, :] target_view = target
    cdef double [:, :, :] source_view = source

    copy_three_dimensions_cython(target_view, source_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void copy_three_dimensions_cython(double[:, :, :] target, double[:, :, :] source, int cores) noexcept:
    cdef long long i, j, k
    cdef long long a = target.shape[0]
    cdef long long b = target.shape[1]
    cdef long long c = target.shape[2]

    for i in prange(a, nogil=True, num_threads=cores):
        for k in range(c):
            for j in range(b):
                target[i, j, k] = source[i, j, k]


def assign_link_loads(actual_links, compressed_links, crosswalk, cores):
    cdef int c = cores

    cdef double [:, :] actual_view = actual_links
    cdef double [:, :] compressed_view = compressed_links
    cdef long long [:] crosswalk_view = crosswalk

    assign_link_loads_cython(actual_view, compressed_view, crosswalk_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void assign_link_loads_cython(
    cython.floating[:, :] actual,
    cython.floating[:, :] compressed,
    long long[:] crosswalk,
    int cores
) noexcept:
    cdef long long i, j, k
    cdef long long links = actual.shape[0]
    cdef long long n = actual.shape[1]

    for i in prange(links, nogil=True, num_threads=cores):
        for j in range(n):
            k = crosswalk[i]
            actual[i, j] = compressed[k, j]


def aggregate_link_costs(actual_costs, compressed_costs, crosswalk):
    cdef double [:] actual_view = actual_costs
    cdef double [:] compressed_view = compressed_costs
    cdef long long [:] crosswalk_view = crosswalk

    aggregate_link_costs_cython(actual_view, compressed_view, crosswalk_view)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void aggregate_link_costs_cython(double[:] actual, double[:] compressed, long long[:] crosswalk) noexcept:
    cdef long long i, k
    cdef long long links = actual.shape[0]
    cdef long long c_l = compressed.shape[0]

    for i in range(c_l):
        compressed[i] = 0

    for i in range(links):
        k = crosswalk[i]
        if k < c_l:
            compressed[k] += actual[i]



import math
from typing import Dict, Optional

import numpy as np
import pandas as pd
import shapely.wkb
import shapely.wkt
from shapely.geometry import Polygon, box
from shapely import union_all

from aequilibrae.context import get_logger
from aequilibrae.parameters import Parameters
from aequilibrae.project.network.gmns_builder import GMNSBuilder
from aequilibrae.project.network.gmns_exporter import GMNSExporter
from aequilibrae.project.network.haversine import haversine
from aequilibrae.project.network.link_types import LinkTypes
from aequilibrae.project.network.links import Links
from aequilibrae.project.network.modes import Modes
from aequilibrae.project.network.nodes import Nodes
from aequilibrae.project.network.osm.osm_builder import OSMBuilder
from aequilibrae.project.network.osm.osm_downloader import OSMDownloader
from aequilibrae.project.network.osm.place_getter import placegetter
from aequilibrae.project.network.periods import Periods
from aequilibrae.project.project_creation import req_link_flds, req_node_flds, protected_fields
from aequilibrae.utils.aeq_signal import SIGNAL
from aequilibrae.utils.interface.worker_thread import WorkerThread


class Network(WorkerThread):
    """
    Network class. Member of an AequilibraE Project
    """

    req_link_flds = req_link_flds
    req_node_flds = req_node_flds
    protected_fields = protected_fields
    link_types: LinkTypes = None
    signal = SIGNAL(object)

    def __init__(self, project) -> None:
        WorkerThread.__init__(self, None)
        from aequilibrae.paths import Graph

        self.graphs = {}  # type: Dict[Graph]
        self.project = project
        self.logger = project.logger
        self.modes = Modes(self)
        self.link_types = LinkTypes(self)
        self.links = Links(self)
        self.nodes = Nodes(self)
        self.periods = Periods(self)

    def skimmable_fields(self):
        """
        Returns a list of all fields that can be skimmed

        :Returns:
            :obj:`list`: List of all fields that can be skimmed
        """

        with self.project.db_connection as conn:
            field_names = conn.execute("PRAGMA table_info(links);").fetchall()

        ignore_fields = ["ogc_fid", "geometry"] + self.req_link_flds
        skimmable = [
            "INT",
            "INTEGER",
            "TINYINT",
            "SMALLINT",
            "MEDIUMINT",
            "BIGINT",
            "UNSIGNED BIG INT",
            "INT2",
            "INT8",
            "REAL",
            "DOUBLE",
            "DOUBLE PRECISION",
            "FLOAT",
            "DECIMAL",
            "NUMERIC",
        ]
        all_fields = []

        for f in field_names:
            if f[1] in ignore_fields:
                continue
            for i in skimmable:
                if i in f[2].upper():
                    all_fields.append(f[1])
                    break

        all_fields.append("distance")
        real_fields = []
        for f in all_fields:
            if f[-2:] == "ab":
                if f[:-2] + "ba" in all_fields:
                    real_fields.append(f[:-3])
            elif f[-3:] != "_ba":
                real_fields.append(f)

        return real_fields

    def list_modes(self):
        """
        Returns a list of all the modes in this model

        :Returns:
            :obj:`list`: List of all modes
        """

        with self.project.db_connection as conn:
            all_modes = [x[0] for x in conn.execute("""select mode_id from modes""").fetchall()]
        return all_modes

    def create_from_osm(
        self,
        model_area: Optional[Polygon] = None,
        place_name: Optional[str] = None,
        modes=("car", "transit", "bicycle", "walk"),
        clean=True,
    ) -> None:
        """
        Downloads the network from OpenStreetMap (OSM)

        :Arguments:
            **area** (:obj:`Polygon`, *Optional*): Polygon for which the network will be downloaded. If not provided,
            a place name would be required

            **place_name** (:obj:`str`, *Optional*): If not downloading with East-West-North-South boundingbox, this is
            required

            **modes** (:obj:`tuple`, *Optional*): List of all modes to be downloaded. Defaults to the modes in the
            parameter file

            **clean** (:obj:`bool`, *Optional*): Keeps only the links that intersects the model area polygon.
            Defaults to ``True``. Does not apply to networks downloaded with a place name

        .. code-block:: python

            >>> project = Project()
            >>> project.new(project_path)

            # Now we can import the network for any place we want
            >>> project.network.create_from_osm(place_name="my_beautiful_hometown") # doctest: +SKIP

            >>> project.close()
        """

        if self.count_links() > 0:
            raise FileExistsError("You can only import an OSM network into a brand new model file")

        with self.project.db_connection as conn:
            conn.execute("""ALTER TABLE links ADD COLUMN osm_id integer""")
            conn.execute("""ALTER TABLE nodes ADD COLUMN osm_id integer""")

        if isinstance(modes, (tuple, list)):
            modes = list(modes)
        elif isinstance(modes, str):
            modes = [modes]
        else:
            raise ValueError("'modes' needs to be string or list/tuple of string")

        if place_name is None:
            if (
                model_area.bounds[0] < -180
                or model_area.bounds[2] > 180
                or model_area.bounds[1] < -90
                or model_area.bounds[3] > 90
            ):
                raise ValueError("Coordinates out of bounds. Polygon must be in WGS84")
            west, south, east, north = model_area.bounds
        else:
            clean = False
            bbox, report = placegetter(place_name)
            if bbox is None:
                msg = f'We could not find a reference for place name "{place_name}"'
                self.logger.warning(msg)
                return
            for i in report:
                if "PLACE FOUND" in i:
                    self.logger.info(i)
            model_area = box(*bbox)
            west, south, east, north = bbox

        # Need to compute the size of the bounding box to not exceed it too much
        height = haversine((east + west) / 2, south, (east + west) / 2, north)
        width = haversine(east, (north + south) / 2, west, (north + south) / 2)
        area = height * width

        par = Parameters().parameters["osm"]
        max_query_area_size = par["max_query_area_size"]

        if area < max_query_area_size:
            polygons = [model_area]
        else:
            polygons = []
            parts = math.ceil(area / max_query_area_size)
            horizontal = math.ceil(math.sqrt(parts))
            vertical = math.ceil(parts / horizontal)
            dx = (east - west) / horizontal
            dy = (north - south) / vertical
            for i in range(horizontal):
                xmin = max(-180, west + i * dx)
                xmax = min(180, west + (i + 1) * dx)
                for j in range(vertical):
                    ymin = max(-90, south + j * dy)
                    ymax = min(90, south + (j + 1) * dy)
                    subarea = box(xmin, ymin, xmax, ymax)
                    if subarea.intersects(model_area):
                        polygons.append(subarea)
        self.logger.info("Downloading data")
        dwnloader = OSMDownloader(polygons, modes, logger=self.logger)
        dwnloader.signal = self.signal
        dwnloader.doWork()

        self.logger.info("Building Network")
        self.builder = OSMBuilder(dwnloader.data, project=self.project, model_area=model_area, clean=clean)

        self.builder.signal = self.signal
        self.builder.doWork()

        self.logger.info("Network built successfully")

    def create_from_gmns(
        self,
        link_file_path: str,
        node_file_path: str,
        use_group_path: str = None,
        geometry_path: str = None,
        srid: int = 4326,
    ) -> None:
        """
        Creates AequilibraE model from links and nodes in GMNS format.

        :Arguments:
            **link_file_path** (:obj:`str`): Path to a links csv file in GMNS format

            **node_file_path** (:obj:`str`): Path to a nodes csv file in GMNS format

            **use_group_path** (:obj:`str`, *Optional*): Path to a csv table containing groupings of uses.
            This helps AequilibraE know when a GMNS use is actually a group of other GMNS uses

            **geometry_path** (:obj:`str`, *Optional*): Path to a csv file containing geometry information for a line
            object, if not specified in the link table

            **srid** (:obj:`int`, *Optional*): Spatial Reference ID in which the GMNS geometries were created
        """

        gmns_builder = GMNSBuilder(self, link_file_path, node_file_path, use_group_path, geometry_path, srid)
        gmns_builder.doWork()

        self.logger.info("Network built successfully")

    def export_to_gmns(self, path: str):
        """
        Exports AequilibraE network to csv files in GMNS format.

        :Arguments:
            **path** (:obj:`str`): Output folder path.
        """

        gmns_exporter = GMNSExporter(self, path)
        gmns_exporter.doWork()

        self.logger.info("Network exported successfully")

    def build_graphs(self, fields: list = None, modes: list = None, limit_to_area: Polygon = None) -> None:
        """Builds graphs for all modes currently available in the model

        When called, it overwrites all graphs previously created and stored in the networks'
        dictionary of graphs

        :Arguments:
            **fields** (:obj:`list`, *Optional*): When working with very large graphs with large number of fields in the
            database, it may be useful to specify which fields to use

            **modes** (:obj:`list`, *Optional*): When working with very large graphs with large number of fields in the
            database, it may be useful to generate only those we need

            **limit_to_area** (:obj:`Polygon`, *Optional*): When working with a very large model area, you may want to
            filter your database to a small area for your computation, which you can do by providing a polygon.
            The search is limited to a spatial index search, so it is very fast but NOT PRECISE.

        To use the 'fields' parameter, a minimalistic option is the following

        .. code-block:: python

            >>> project = create_example(project_path)

            >>> fields = ['distance']
            >>> project.network.build_graphs(fields, modes = ['c', 'w'])

        """
        from aequilibrae.paths import Graph

        with self.project.db_connection as conn:
            if fields is None:
                field_names = conn.execute("PRAGMA table_info(links);").fetchall()

                ignore_fields = ["ogc_fid", "geometry"]
                all_fields = [f[1] for f in field_names if f[1] not in ignore_fields]
            else:
                fields.extend(["link_id", "a_node", "b_node", "direction", "modes"])
                all_fields = list(set(fields))

            if modes is None:
                modes = conn.execute("select mode_id from modes;").fetchall()
                modes = [m[0] for m in modes]
            elif isinstance(modes, str):
                modes = [modes]

            if limit_to_area is not None:
                spatial_add = """ WHERE links.rowid in (
                                        select rowid from SpatialIndex where f_table_name = 'links' and
                                       search_frame = GeomFromWKB(?, 4326))"""

            sql = f"select {','.join(all_fields)} from links"

            sql_centroids = "select node_id from nodes where is_centroid=1 order by node_id;"
            centroids = np.array([i[0] for i in conn.execute(sql_centroids).fetchall()], np.uint32)
            centroids = centroids if centroids.shape[0] else None

            with pd.option_context("future.no_silent_downcasting", True):
                if limit_to_area is None:
                    df = pd.read_sql(sql, conn).fillna(value=np.nan).infer_objects(False)
                else:
                    sql += spatial_add
                    df = (
                        pd.read_sql_query(sql, conn, params=(limit_to_area.wkb,))
                        .fillna(value=np.nan)
                        .infer_objects(False)
                    )

                    # We filter to centroids existing in our filtered area
                    centroids = centroids[np.isin(centroids, df.a_node) | np.isin(centroids, df.b_node)]

            valid_fields = list(df.select_dtypes(np.number).columns) + ["modes"]

        lonlat = self.nodes.lonlat.set_index("node_id")
        data = df[valid_fields]
        for m in modes:
            # For any link in net that doesn't support mode 'm', set a_node = b_node (these will be culled when
            # the compressed graph representation is created)
            net = pd.DataFrame(data, copy=True)
            net.loc[~net.modes.str.contains(m), "b_node"] = net.loc[~net.modes.str.contains(m), "a_node"]

            g = Graph()
            g.mode = m
            g.network = net
            g.prepare_graph(centroids)
            g.set_blocked_centroid_flows(True)
            if centroids is None:
                get_logger().warning("Your graph has no centroids")
            g.lonlat_index = lonlat.loc[g.all_nodes]
            self.graphs[m] = g

    def set_time_field(self, time_field: str) -> None:
        """
        Set the time field for all graphs built in the model

        :Arguments:
            **time_field** (:obj:`str`): Network field with travel time information
        """
        for m, g in self.graphs.items():
            if time_field not in list(g.graph.columns):
                raise ValueError(f"{time_field} not available. Check if you have NULL values in the database")
            g.free_flow_time = time_field
            g.set_graph(time_field)
            self.graphs[m] = g

    def count_links(self) -> int:
        """
        Returns the number of links in the model

        :Returns:
            :obj:`int`: Number of links
        """
        return self.__count_items("link_id", "links", "link_id>=0")

    def count_centroids(self) -> int:
        """
        Returns the number of centroids in the model

        :Returns:
            :obj:`int`: Number of centroids
        """
        return self.__count_items("node_id", "nodes", "is_centroid=1")

    def count_nodes(self) -> int:
        """
        Returns the number of nodes in the model

        :Returns:
            :obj:`int`: Number of nodes
        """
        return self.__count_items("node_id", "nodes", "node_id>=0")

    def extent(self):
        """Queries the extent of the network included in the model

        :Returns:
            **model extent** (:obj:`Polygon`): Shapely polygon with the bounding box of the model network.
        """
        with self.project.db_connection as conn:
            poly = shapely.wkb.loads(conn.execute('Select ST_asBinary(GetLayerExtent("Links"))').fetchone()[0])
        return poly

    def convex_hull(self) -> Polygon:
        """Queries the model for the convex hull of the entire network

        :Returns:
            **model coverage** (:obj:`Polygon`): Shapely (Multi)polygon of the model network.
        """
        with self.project.db_connection as conn:
            sql = 'Select ST_asBinary("geometry") from Links where ST_Length("geometry") > 0;'
            links = [shapely.wkb.loads(x[0]) for x in conn.execute(sql).fetchall()]
        return union_all(links).convex_hull

    def __count_items(self, field: str, table: str, condition: str) -> int:
        with self.project.db_connection as conn:
            c = conn.execute(f"select count({field}) from {table} where {condition};").fetchone()[0]
        return c



import pathlib
import sqlite3
from dataclasses import dataclass
from enum import IntEnum
from typing import Optional

from aequilibrae import logger
from aequilibrae.utils.model_run_utils import import_file_as_module


class MigrationStatus(IntEnum):
    MISSING: int = 1
    SKIPPED: int = 2
    APPLIED: int = 3


@dataclass
class Migration:
    """
    Small utility class to wrap files used for database upgrades/migrations.

    Individual migrations can report their status, be marked as 'seen' or as another status, and applied. SQL migrations
    are executed using ``sqlite3.executescript``. Python migrations are loaded as a module, they should expose a
    ``migrate`` function which accepts an ``sqlite3.Connection`` as a single positional argument.

    Marking a migration as 'seen' will add it to the ``migrations`` table as ``MISSING`` if it is not already
    present. If it is present no change is made.

    Applying a migration will update the status to 'APPLIED' with the current timestamp.

    A migration's status cannot be downgraded without force.

    Migrations are identified based on their ``id`` attribute and the ``id`` field of the ``migrations`` table.
    """

    id: int
    name: str
    file: pathlib.Path
    type: str = None

    def __post_init__(self):
        if self.file.suffix == ".py":
            self.type = "py"
        elif self.file.suffix == ".sql":
            self.type = "sql"
        else:
            raise ValueError("only Python ('.py') and SQL ('.sql') files are supported for migrations")

    def status(self, conn: sqlite3.Connection) -> MigrationStatus:
        """
        Query the database for this migrations status.

        If the ``migrations`` table is not present all migrations are considered ``MISSING``.

        :Arguments:
            **conn** (:obj:`sqlite3.Connection`): SQLite database connection.

        :Returns:
            **status** (:obj:`MigrationStatus`): Migration status enum.
        """
        res = conn.execute("SELECT status FROM migrations WHERE id=?", (self.id,)).fetchone()
        return MigrationStatus.MISSING if res is None else MigrationStatus[res[0]]

    def mark_as(self, conn: sqlite3.Connection, status: MigrationStatus, force: bool = False):
        """
        Update or insert this migration with the given status.

        If the migration is not present in the table it will be inserted. If it is present and the new status is a
        'upgrade' or ``force=True``, then it will be updated. Otherwise no change will be made.

        :Arguments:
            **conn** (:obj:`sqlite3.Connection`): SQLite database connection.
            **status** (:obj:`MigrationStatus`): Migration status enum.
        """
        res = conn.execute("SELECT status FROM migrations WHERE id=?", (self.id,)).fetchone()
        if res is None:
            conn.execute(
                "INSERT INTO migrations (id, name, status, date) VALUES(?,?,?,CURRENT_TIMESTAMP)",
                (self.id, self.name, status.name),
            )
        else:
            res = MigrationStatus[res[0]]
            if force or res < status or res < status < MigrationStatus.APPLIED:
                # We want to allow marking the status as APPLIED if it is MISSING or SKIPPED, and as SKIPPED if it
                # is MISSING, or just whenever force is True
                conn.execute(
                    "UPDATE migrations SET status=?, name=?, date=CURRENT_TIMESTAMP WHERE id=?",
                    (status.name, self.name, self.id),
                )

    def mark_as_seen(self, conn: sqlite3.Connection):
        """
        Mark this migration as 'seen'.

        Marking a migration as 'seen' will add it to the ``migrations`` table as ``MISSING`` if it is not already
        present. If it is present no change is made.

        :Arguments:
            **conn** (:obj:`sqlite3.Connection`): SQLite database connection.
        """
        self.mark_as(conn, MigrationStatus.MISSING, force=False)

    def apply(self, conn: sqlite3.Connection):
        """
        Apply this migration.

        Successful application will mark the migration as ``APPLIED``.

        Python migrations should never use ``executescript`` as it will commit the pending transaction and place SQLite
        in autocommit mode. If the migration then fails the database will be bad state.

        :Arguments:
            **conn** (:obj:`sqlite3.Connection`): SQLite database connection.
        """
        if self.type == "py":
            self._apply_python(conn)
        elif self.type == "sql":
            self._apply_sql(conn)
        else:
            raise ValueError("only Python ('.py') and SQL ('.sql') files are supported for migrations")

        self.mark_as(conn, MigrationStatus.APPLIED)
        logger.info(f"Completed migration '{self.name}'")

    def _apply_sql(self, conn: sqlite3.Connection):
        with open(self.file, "r") as f:
            contents = f.read()
        conn.executescript(contents)

    def _apply_python(self, conn: sqlite3.Connection):
        module = import_file_as_module(self.file, self.name)
        try:
            migrate = module.migrate
        except AttributeError as e:
            raise RuntimeError(f"'{self.name} does not expose a global 'migrate' callable") from e

        if not callable(migrate):
            raise RuntimeError("found 'migrate' symbol in the migration file but it is not callable")

        migrate(conn)


class MigrationManager:
    """
    Small utility class to manage, validate, and apply a set of ``Migration``s.

    :Arguments:
        **migration_file** (:obj:`pathlib.Path`): A path to a Python with which defines a global ``migrations`` variable
            as a list of ``pathlib.Path`` to migrations.
    """

    network_migration_file = (
        pathlib.Path(__file__).parent.parent / "database_specification" / "network" / "migrations" / "migrations.py"
    )
    transit_migration_file = (
        pathlib.Path(__file__).parent.parent / "database_specification" / "transit" / "migrations" / "migrations.py"
    )

    def __init__(self, migration_file: pathlib.Path):
        migrations = import_file_as_module(
            migration_file,
            "aequilibrae.project.database_specification.migrations",
        ).migrations

        res = []
        for migration in migrations:
            if not migration.exists():
                raise FileNotFoundError(f"migration file '{migration.name}' does not exist'")

            id, _, name = migration.stem.partition("_")
            id = int(id)
            if id < 0:
                raise ValueError("migration IDs must be >= 0")
            res.append(Migration(id=id, name=name, file=migration))

        self.migrations: dict[int, Migration] = {
            migration.id: migration for migration in sorted(res, key=lambda x: x.id)
        }
        if len(self.migrations) != len(res):
            raise ValueError("duplicate migration IDs found. Ensure migration IDs are unique.")

    def __ensure_inital_is_applied(self, conn):
        # Handle the initial migration separately, the 'migrations' table might not have been created. We implicitly
        # apply this migration all the time to ensure the table exists.
        self.migrations[0].apply(conn)

    def status(self, conn: sqlite3.Connection) -> dict[int, MigrationStatus]:
        """
        Query the database for all migrations' status.

        If the ``migrations`` table is not present all migrations are considered ``MISSING``.

        :Arguments:
            **conn** (:obj:`sqlite3.Connection`): SQLite database connection.

        :Returns:
            **status** (:obj:`dict[int, MigrationStatus]`): Migration status enums by their ID.
        """
        self.__ensure_inital_is_applied(conn)
        return {k: v.status(conn) for k, v in self.migrations.items()}

    def mark_all_as_seen(self, conn: sqlite3.Connection):
        """
        Mark all migrations as 'seen'.

        Marking a migration as 'seen' will add it to the ``migrations`` table as ``MISSING`` if it is not already
        present. If it is present no change is made.

        :Arguments:
            **conn** (:obj:`sqlite3.Connection`): SQLite database connection.
        """
        self.__ensure_inital_is_applied(conn)
        for migration in self.migrations.values():
            migration.mark_as_seen(conn)

    def find_applicable(self, conn: sqlite3.Connection):
        """
        Find all applicable migrations.

        A migration is applicable if all migrations before it (ordered by ID) have been applied or skipped.

        If an out-of-order migration is detected a ``RuntimeError` will raised and manual intervention will be required.

        :Arguments:
            **conn** (:obj:`sqlite3.Connection`): SQLite database connection.
        """
        migrations = list(self.status(conn).items())

        for i in range(len(migrations)):
            k, v = migrations[i]
            if v == MigrationStatus.MISSING:
                break
        else:
            i += 1

        res = []
        for j in range(i, len(migrations)):
            k, v = migrations[j]
            if v == MigrationStatus.APPLIED:
                raise RuntimeError("out of order migration application found. Manual intervention required")
            else:
                res.append(self.migrations[k])

        return res

    def upgrade(self, conn: sqlite3.Connection, skip: set[int] = None):
        """
        Find and apply all applicable migrations.

        Optionally skip some migrations. Take care when skipping migrations.

        :Arguments:
            **conn** (:obj:`sqlite3.Connection`): SQLite database connection.
            **skip** (:obj:`set[int]`): Set of migration IDs to skip.
        """
        if skip is None:
            skip = set()
        migrations = self.find_applicable(conn)

        iso_lvl = conn.isolation_level
        conn.isolation_level = None
        try:
            for migration in migrations:
                conn.execute("BEGIN")
                with conn:
                    if migration.id in skip:
                        migration.mark_as(conn, MigrationStatus.SKIPPED)
                    else:
                        migration.apply(conn)
        finally:
            conn.isolation_level = iso_lvl



from libcpp.vector cimport vector
from libcpp.memory cimport unique_ptr, make_unique
from libcpp cimport bool

cdef class Sparse:
    pass

cdef struct COO_f64_struct:
    unique_ptr[vector[size_t]] row
    unique_ptr[vector[size_t]] col
    unique_ptr[vector[double]] f64_data

cdef struct COO_f32_struct:
    unique_ptr[vector[size_t]] row
    unique_ptr[vector[size_t]] col
    unique_ptr[vector[float]] f32_data

cdef class COO(Sparse):
    cdef:
        unique_ptr[vector[size_t]] row
        unique_ptr[vector[size_t]] col
        bool f64
        unique_ptr[vector[double]] f64_data
        unique_ptr[vector[float]] f32_data
        public object shape

    cdef void f64_append(COO self, size_t i, size_t j, double v) noexcept nogil
    cdef void f32_append(COO self, size_t i, size_t j, float v) noexcept nogil

    @staticmethod
    cdef void init_f64_struct(COO_f64_struct &struct) noexcept nogil

    @staticmethod
    cdef void init_f32_struct(COO_f32_struct &struct) noexcept nogil

    @staticmethod
    cdef object from_f64_struct(COO_f64_struct &struct)

    @staticmethod
    cdef object from_f32_struct(COO_f32_struct &struct)

    @staticmethod
    cdef void f64_struct_append(COO_f64_struct &struct, size_t i, size_t j, double v) noexcept nogil

    @staticmethod
    cdef void f32_struct_append(COO_f32_struct &struct, size_t i, size_t j, float v) noexcept nogil



# cython: boundscheck=False, wraparound=False, embedsignature=False, cdivision=True, initializedcheck=False

""" Priority queue based on a minimum 4-ary heap.

    4-ary heap implemented with a static array.

    Tree elements also stored in a static array.

author : François Pacull
email: francois.pacull@architecture-performance.fr
copyright : Architecture & Performance
license :

MIT License

Copyright (c) 2022 Architecture & Performance

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from libc.stdlib cimport malloc, free
from libc.math cimport INFINITY
include "parameters.pxi"

cdef enum ElementState:
    SCANNED
    NOT_IN_HEAP
    IN_HEAP

cdef struct Element:
    DTYPE_t key
    ElementState state
    size_t node_idx

cdef struct PriorityQueue:
    size_t length  # number of elements in the array
    size_t size  # number of elements in the heap
    size_t* A  # array storing the binary tree
    Element* Elements  # array storing the elements
    DTYPE_t* keys

cdef void init_heap(PriorityQueue* pqueue, size_t length) noexcept nogil:
    """Initialize the binary heap.

    input
    =====
    * PriorityQueue* pqueue : 4-ary heap based priority queue
    * size_t length : length (maximum size) of the heap
    """
    cdef size_t i

    pqueue.length = length
    pqueue.size = 0
    pqueue.A = <size_t*> malloc(length * sizeof(size_t))
    pqueue.Elements = <Element*> malloc(length * sizeof(Element))

    for i in range(length):
        pqueue.A[i] = length
        _initialize_element(pqueue, i)


cdef void _initialize_element(PriorityQueue* pqueue, size_t element_idx) noexcept nogil:
    """Initialize a single element.

    input
    =====
    * PriorityQueue* pqueue : 4-ary heap based priority queue
    * size_t element_idx : index of the element in the element array
    """
    pqueue.Elements[element_idx].key = INFINITY
    pqueue.Elements[element_idx].state = NOT_IN_HEAP
    pqueue.Elements[element_idx].node_idx = pqueue.length


cdef void free_heap(PriorityQueue* pqueue) noexcept nogil:
    """Free the binary heap.

    input
    =====
    * PriorityQueue* pqueue : 4-ary heap based priority queue
    """
    free(pqueue.A)
    free(pqueue.Elements)


cdef void insert(PriorityQueue* pqueue, size_t element_idx, DTYPE_t key) noexcept nogil:
    """Insert an element into the heap and reorder the heap.

    input
    =====
    * PriorityQueue* pqueue : 4-ary heap based priority queue
    * size_t element_idx : index of the element in the element array
    * DTYPE_t key : key value of the element

    assumptions
    ===========
    * the element pqueue.Elements[element_idx] is not in the heap
    * its new key is smaller than INFINITY
    """
    cdef size_t node_idx = pqueue.size

    pqueue.size += 1
    pqueue.Elements[element_idx].state = IN_HEAP
    pqueue.Elements[element_idx].node_idx = node_idx
    pqueue.A[node_idx] = element_idx
    _decrease_key_from_node_index(pqueue, node_idx, key)


cdef void decrease_key(PriorityQueue* pqueue, size_t element_idx, DTYPE_t key_new) noexcept nogil:
    """Decrease the key of a element in the heap, given its element index.

    input
    =====
    * PriorityQueue* pqueue : 4-ary heap based priority queue
    * size_t element_idx : index of the element in the element array
    * DTYPE_t key_new : new value of the element key

    assumption
    ==========
    * pqueue.Elements[idx] is in the heap
    """
    _decrease_key_from_node_index(
        pqueue,
        pqueue.Elements[element_idx].node_idx,
        key_new)


cdef DTYPE_t peek(PriorityQueue* pqueue) noexcept nogil:
    """Find heap min key.

    input
    =====
    * PriorityQueue* pqueue : 4-ary heap based priority queue

    output
    ======
    * DTYPE_t : key value of the min element

    assumption
    ==========
    * pqueue.size > 0
    * heap is heapified
    """
    return pqueue.Elements[pqueue.A[0]].key


cdef bint is_empty(PriorityQueue* pqueue) noexcept nogil:
    """Check whether the heap is empty.

    input
    =====
    * PriorityQueue* pqueue : 4-ary heap based priority queue
    """
    cdef bint isempty = 0

    if pqueue.size == 0:
        isempty = 1

    return isempty


cdef size_t extract_min(PriorityQueue* pqueue) noexcept nogil:
    """Extract element with min keay from the heap,
    and return its element index.

    input
    =====
    * PriorityQueue* pqueue : 4-ary heap based priority queue

    output
    ======
    * size_t : element index with min key

    assumption
    ==========
    * pqueue.size > 0
    """
    cdef:
        size_t element_idx = pqueue.A[0]  # min element index
        size_t node_idx = pqueue.size - 1  # last leaf node index

    # exchange the root node with the last leaf node
    _exchange_nodes(pqueue, 0, node_idx)

    # remove this element from the heap
    pqueue.Elements[element_idx].state = SCANNED
    pqueue.Elements[element_idx].node_idx = pqueue.length
    pqueue.A[node_idx] = pqueue.length
    pqueue.size -= 1

    # reorder the tree Elements from the root node
    _min_heapify(pqueue, 0)

    return element_idx

cdef void _exchange_nodes(PriorityQueue* pqueue, size_t node_i, size_t node_j) noexcept nogil:
    """Exchange two nodes in the heap.

    input
    =====
    * PriorityQueue* pqueue: binary heap
    * size_t node_i: first node index
    * size_t node_j: second node index
    """
    cdef:
        size_t element_i = pqueue.A[node_i]
        size_t element_j = pqueue.A[node_j]

    # exchange element indices in the heap array
    pqueue.A[node_i] = element_j
    pqueue.A[node_j] = element_i

    # exchange node indices in the element array
    pqueue.Elements[element_j].node_idx = node_i
    pqueue.Elements[element_i].node_idx = node_j


cdef void _min_heapify(PriorityQueue* pqueue, size_t node_idx) noexcept nogil:
    """Re-order sub-tree under a given node (given its node index)
    until it satisfies the heap property.

    input
    =====
    * PriorityQueue* pqueue : 4-ary heap based priority queue
    * size_t node_idx : node index
    """
    cdef:
        size_t c1, c2, c3, c4, i = node_idx, s
        DTYPE_t val_tmp, val_min

    if pqueue.size == 0:
        return

    while True:

        c1 = 4 * i + 1
        c2 = c1 + 1
        c3 = c2 + 1
        c4 = c3 + 1

        s = i
        val_min = pqueue.Elements[pqueue.A[s]].key
        if (c4 < pqueue.size):
            val_tmp = pqueue.Elements[pqueue.A[c4]].key
            if val_tmp < val_min:
                s = c4
                val_min = val_tmp
            val_tmp = pqueue.Elements[pqueue.A[c3]].key
            if val_tmp < val_min:
                s = c3
                val_min = val_tmp
            val_tmp = pqueue.Elements[pqueue.A[c2]].key
            if val_tmp < val_min:
                s = c2
                val_min = val_tmp
            val_tmp = pqueue.Elements[pqueue.A[c1]].key
            if val_tmp < val_min:
                s = c1
        else:
            if (c3 < pqueue.size):
                val_tmp = pqueue.Elements[pqueue.A[c3]].key
                if val_tmp < val_min:
                    s = c3
                    val_min = val_tmp
                val_tmp = pqueue.Elements[pqueue.A[c2]].key
                if val_tmp < val_min:
                    s = c2
                    val_min = val_tmp
                val_tmp = pqueue.Elements[pqueue.A[c1]].key
                if val_tmp < val_min:
                    s = c1
            else:
                if (c2 < pqueue.size):
                    val_tmp = pqueue.Elements[pqueue.A[c2]].key
                    if val_tmp < val_min:
                        s = c2
                        val_min = val_tmp
                    val_tmp = pqueue.Elements[pqueue.A[c1]].key
                    if val_tmp < val_min:
                        s = c1
                else:
                    if (c1 < pqueue.size):
                        val_tmp = pqueue.Elements[pqueue.A[c1]].key
                        if val_tmp < val_min:
                            s = c1

        if s != i:
            _exchange_nodes(pqueue, i, s)
            i = s
        else:
            break


cdef void _decrease_key_from_node_index(PriorityQueue* pqueue, size_t node_idx, DTYPE_t key_new) noexcept nogil:
    """Decrease the key of an element in the heap, given its tree index.

    input
    =====
    * PriorityQueue* pqueue : 4-ary heap based priority queue
    * size_t node_idx : node index
    * DTYPE_t key_new : new key value

    assumptions
    ===========
    * pqueue.Elements[pqueue.A[node_idx]] is in the heap (node_idx < pqueue.size)
    * key_new < pqueue.Elements[pqueue.A[node_idx]].key
    """
    cdef:
        size_t i = node_idx, j
        DTYPE_t key_j

    pqueue.Elements[pqueue.A[i]].key = key_new
    while i > 0:
        j = (i - 1) // 4
        key_j = pqueue.Elements[pqueue.A[j]].key
        if key_j > key_new:
            _exchange_nodes(pqueue, i, j)
            i = j
        else:
            break



# This is a convenience file to allow developers easier compilation of just the assignment module
# Generally compilation is done directly from the root setup.py
# Make sure to keep this up to date with that file

import platform
import numpy as np
from Cython.Build import cythonize

try:
    from setuptools import setup
    from setuptools import Extension
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension

is_win = "WINDOWS" in platform.platform().upper()
prefix = "/" if is_win else "-f"
cpp_std = "/std:c++17" if is_win else "-std=c++17"

ext_modules = [
    Extension(
        "AoN",
        ["AoN.pyx"],
        extra_compile_args=[f"{prefix}openmp", cpp_std],  # do we want -Ofast?
        extra_link_args=[f"{prefix}openmp"],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
        include_dirs=[np.get_include()],
    )
]

setup(name="AoN", ext_modules=cythonize(ext_modules))



from os.path import dirname, abspath, join

spatialite_database = join(dirname(dirname(abspath(__file__))), "reference_files", "spatialite.sqlite")
demo_init_py = join(dirname(dirname(abspath(__file__))), "reference_files", "demo.__init__.py")



def clean(project):
    # Since we cannot decide the order of trigger execution in SQLITE, we make sure to remove any
    # extraneous nodes at a few key moments (i.e. opening and closing the model)
    with project.db_connection as conn:

        sqls = [
            """DELETE from Nodes where is_centroid=0 and
                                        (SELECT count(*) FROM links WHERE a_node = node_id OR b_node = node_id) = 0;"""
        ]

        for sql in sqls:
            conn.execute(sql)



import warnings
from copy import deepcopy
from math import ceil
from typing import List

import numpy as np
import pandas as pd
from shapely.geometry.linestring import LineString
from shapely.ops import linemerge
from shapely.ops import substring

from aequilibrae.context import get_active_project
from aequilibrae.paths.graph import Graph
from aequilibrae.utils.aeq_signal import SIGNAL
from aequilibrae.utils.db_utils import commit_and_close
from aequilibrae.utils.interface.worker_thread import WorkerThread


class NetworkSimplifier(WorkerThread):
    signal = SIGNAL(object)

    def __init__(self, project=None) -> None:
        super().__init__(None)

        self.project = project or get_active_project()
        self.network = self.project.network
        self.link_layer = self.network.links.data

        warnings.warn("This will alter your database in place. Make sure you have a backup.")

    def simplify(self, graph: Graph, max_speed_ratio: float = 1.1):
        """
        Simplifies the network by merging links that are shorter than a given threshold

        :Arguments:
            **graph** (:obj:`Graph`): AequilibraE graph

            **max_speed_ratio** (:obj:`float`, *Optional*): Maximum ratio between the fastest
            and slowest speed for a link to be considered for simplification.
        """
        self.graph = graph

        # Creates the sequence of compressed link IDs so we don't have to rebuild through path computation
        compressed_idx, compressed_link_data, _ = self.graph.create_compressed_link_network_mapping()

        link_set_df = self.graph.network.merge(self.graph.graph[["link_id", "__compressed_id__"]], on="link_id")

        # compressed_ids that appear for more than one link
        relevant_compressed_ids = pd.DataFrame(link_set_df.value_counts("__compressed_id__")).query("count>1").index
        link_set_df = link_set_df[link_set_df.__compressed_id__.isin(relevant_compressed_ids)]

        # Makes sure we always get the compressed Id for the same (arbitrary) direction
        link_set_df = link_set_df.sort_values("__compressed_id__").drop_duplicates(subset="link_id", keep="first")

        # We only need one link ID per "super link"
        link_set_df = link_set_df.drop_duplicates(subset="__compressed_id__")
        centroid_connectors = self.graph.network.query("link_id not in @self.link_layer.link_id").link_id.to_numpy()

        links_to_delete, new_links = [], []
        max_link_id = self.link_layer.link_id.max() + 1
        self.signal.emit(["start", link_set_df.shape[0], "Simplifying links"])

        counter = 0
        for _, rec in link_set_df.iterrows():
            counter += 1
            self.signal.emit(["update", counter, "Simplifying links"])
            compressed_id = rec.__compressed_id__

            # We got to the group of links where AequilibraE can no longer compress
            if compressed_id + 1 == compressed_idx.shape[0]:
                continue

            link_sequence = compressed_link_data[compressed_idx[compressed_id] : compressed_idx[compressed_id + 1]]
            if len(link_sequence) < 2:
                continue

            link_sequence = [abs(x) for x in link_sequence]
            self.link_sequence = [x for x in link_sequence if x not in centroid_connectors]
            self.candidates = self.link_layer.query("link_id in @link_sequence").set_index("link_id")

            if self.candidates.shape[0] <= 1:
                continue

            # To merge, all links have to have the same number of lanes and link type
            breaker = self.candidates["link_type"].nunique() > 1
            if breaker:
                continue

            # We build the geometry sequence, speeds, and capacities
            candidates, geos, long_dir, long_lnk = self.__process_link_fields(
                self.candidates, self.link_sequence, max_speed_ratio
            )

            if candidates.empty:
                continue

            new_geo = linemerge(geos)
            if not isinstance(new_geo, LineString):
                warnings.warn(f"Failed to merge geometry for superlink around link {rec.link_id}")
                continue

            break_into = ceil(new_geo.length)

            # Now we build the template for the links we will build
            main_data = long_lnk.to_dict()

            # Some values we will bring from the weighted average
            for field in ["speed_ab", "speed_ba", "capacity_ab", "capacity_ba"]:
                metric = (candidates[field] * candidates["distance"]).sum() / candidates["distance"].sum()
                if long_dir == 1:
                    main_data[field] = metric
                else:
                    field2 = field.replace("ab", "ba") if "ab" in field else field.replace("ba", "ab")
                    main_data[field2] = metric

            # If that link is in the opposite direction, we need to swap lanes,
            # as we would have swapped the geometry as well
            if long_dir == -1:
                main_data["lanes_ab"], main_data["lanes_ba"] = main_data["lanes_ba"], main_data["lanes_ab"]

            for i in range(break_into):
                data = deepcopy(main_data)
                data["link_id"] = max_link_id

                sub_geo = substring(new_geo, i / break_into, (i + 1.0) / break_into, normalized=True)
                if sub_geo.length < 0.000001:
                    raise ValueError("Link with zero length")

                data["geo"] = sub_geo.wkb
                max_link_id += 1
                new_links.append(data)
            links_to_delete.extend(candidates.index.tolist())

        self.signal.emit(["finished"])

        self.project.logger.info(f"{len(links_to_delete):,} links will be removed")
        self.project.logger.info(f"{len(new_links):,} links will be added")
        if new_links:
            self.__execute_link_deletion_and_addition(new_links, links_to_delete)

        self.project.logger.warning("Network has been rebuilt. You should run this tool's rebuild network method")

    def __process_link_fields(self, candidates, link_sequence, max_speed_ratio):
        start_node = candidates.loc[link_sequence[0]]["a_node"]
        longest_link_id = candidates.sort_values("distance", ascending=False).index[0]
        speed_ab, speed_ba, geos, lanes_ab, lanes_ba = [], [], [], [], []
        longest_link, longest_direction = None, None
        for link_id in link_sequence:
            link = candidates.loc[link_id]
            direction = "AB" if start_node == link.a_node else "BA"
            geos.append(link.geometry if direction == "AB" else link.geometry.reverse())
            speed_ab.append(link.speed_ab if direction == "AB" else link.speed_ba)
            speed_ba.append(link.speed_ba if direction == "AB" else link.speed_ab)
            lanes_ab.append(link.lanes_ab if direction == "AB" else link.lanes_ba)
            lanes_ba.append(link.lanes_ba if direction == "AB" else link.lanes_ab)
            if link_id == longest_link_id:
                # We use the longest link as a template
                longest_direction = 1 if direction == "AB" else -1
                longest_link = link
            start_node = link.b_node if direction == "AB" else link.a_node

        constraints_broken = int(np.unique(lanes_ab).shape[0] > 1)
        constraints_broken += np.unique(lanes_ba).shape[0] > 1

        # Speeds cannot diverge by more than 10%
        constraints_broken += max(speed_ab) / max(min(speed_ab), 0.00001) > max_speed_ratio
        constraints_broken += max(speed_ba) / max(min(speed_ba), 0.00001) > max_speed_ratio
        if constraints_broken > 0:
            return pd.DataFrame([]), None, None, None

        return candidates, geos, longest_direction, longest_link

    def __execute_link_deletion_and_addition(self, new_links, links_to_delete):
        df = pd.DataFrame(new_links)
        df = df.drop(columns=["a_node", "b_node", "geometry", "ogc_fid"]).rename({"geo": "geometry"}, axis=1)
        cols = list(df.columns)
        df = df[cols]
        data = df.assign(srid=self.link_layer.crs.to_epsg()).to_records(index=False)

        sql = f"INSERT INTO links({','.join(df.columns)}) VALUES ({','.join(['?'] * (len(df.columns) - 1))},GeomFromWKB(?, ?))"
        with commit_and_close(self.project.path_to_file, spatial=True) as conn:
            conn.executemany(sql, data)
            conn.executemany("DELETE FROM links WHERE link_id=?", [[x] for x in links_to_delete])
            conn.commit()

        # Validate that we kept distances the same
        old_dist = self.link_layer.geometry.length.sum()
        new_layer = self.network.links
        new_layer.refresh()
        new_dist = new_layer.data.geometry.length.sum()

        self.project.logger.warning(
            f"Old distance: {old_dist}, new distance: {new_dist}. Difference: {old_dist - new_dist}"
        )
        self.link_layer = new_layer.data

    def collapse_links_into_nodes(self, links: List[int]):
        """
        Collapses links into nodes, adjusting the network in the neighborhood.

        :Arguments:
            **links** (:obj:`List[int]`): List containing link IDs to be collapsed.
        """
        srid = self.link_layer.crs.to_epsg()
        target_links = self.link_layer.query("link_id in @links")
        with commit_and_close(self.project.path_to_file, spatial=True) as conn:
            for _, link in target_links.iterrows():
                wkb = link.geometry.interpolate(0.5, normalized=True).wkb
                conn.execute("DELETE FROM links WHERE link_id=?", [link.link_id])
                conn.commit()
                conn.execute("UPDATE nodes SET geometry=GeomFromWKB(?, ?) WHERE node_id=?", [wkb, srid, link.a_node])
                conn.execute("UPDATE nodes SET geometry=GeomFromWKB(?, ?) WHERE node_id=?", [wkb, srid, link.b_node])
                conn.commit()

        self.link_layer = self.network.links.data
        self.project.logger.warning(f"{len(links)} links collapsed into nodes")

    def rebuild_network(self):
        """Rebuilds the network elements that would have to be rebuilt after massive network simplification"""

        self.network.links.refresh()
        self.network.nodes.refresh()

        with commit_and_close(self.project.path_to_file, spatial=True) as conn:
            conn.execute("VACUUM")



# cython: language_level=3str
from aequilibrae.paths.graph import Graph
from aequilibrae.paths.cython.route_choice_types cimport LinkSet_t, minstd_rand, shuffle
from aequilibrae.matrix.coo_demand cimport GeneralisedCOODemand

from cython.operator cimport dereference as d
from cython.parallel cimport parallel, prange, threadid
from libc.limits cimport UINT_MAX
from libc.string cimport memcpy
from libcpp cimport nullptr
from libcpp.algorithm cimport reverse, copy
from libcpp.unordered_set cimport unordered_set
from libcpp.utility cimport pair
from libcpp.vector cimport vector
from libcpp cimport bool
from openmp cimport omp_get_max_threads

from libcpp.memory cimport shared_ptr

from typing import Tuple
import itertools
import warnings

import numpy as np
import pandas as pd


"""This module aims to implemented the BFS-LE algorithm as described in Rieser-Schüssler, Balmer, and Axhausen, 'Route
Choice Sets for Very High-Resolution Data'.  https://doi.org/10.1080/18128602.2012.671383

A rough overview of the algorithm is as follows.
    1. Prepare the initial graph, this is depth 0 with no links removed.
    2. Find a short path, P. If P is not empty add P to the path set.
    3. For all links p in P, remove p from E, compounding with the previously removed links.
    4. De-duplicate the sub-graphs, we only care about unique sub-graphs.
    5. Go to 2.

Details: The general idea of the algorithm is pretty simple, as is the implementation. The caveats here is that there is
a lot of cpp interop and memory management. A description of the purpose of variables is in order:

route_set: See route_choice.pxd for full type signature. It's an unordered set (hash set) of pointers to vectors of link
IDs. It uses a custom hashing function and comparator. The hashing function is defined in a string that in inlined
directly into the output cpp. This is done allow declaring of the `()` operator, which is required and AFAIK not
possible in Cython. The hash is designed to dereference then hash order dependent vectors. One isn't provided by
stdlib. The comparator simply dereferences the pointer and uses the vector comparator. It's designed to store the
outputted paths. Heap allocated (needs to be returned).

removed_links: See route_choice.pxd for full type signature. It's an unordered set of pointers to unordered sets of link
IDs. Similarly to `route_set` is uses a custom hash function and comparator. This hash function is designed to be order
independent and should only use commutative operations. The comparator is the same. It's designed to store all of the
removed link sets we've seen before. This allows us to detected duplicated graphs.

rng: A custom imported version of std::linear_congruential_engine. libcpp doesn't provide one so we do. It should be
significantly faster than the std::mersenne_twister_engine without sacrificing much. We don't need amazing RNG, just ok
and fast. This is only used to shuffle the queue.

queue, next_queue: These are vectors of pointers to sets of removed links. We never need to push to the front of these
so a vector is best. We maintain two queues, one that we are currently iterating over, and one that we can add to,
building up with all the newly removed link sets. These two are swapped at the end of an iteration, next_queue is then
cleared. These store sets of removed links.

banned, next_banned: `banned` is the iterator variable for `queue`. `banned` is copied into `next_banned` where another
link can be added without mutating `banned`. If we've already seen this set of removed links `next_banned` is
immediately deallocated. Otherwise it's placed into `next_queue`.

vec: `vec` is a scratch variable to store pointers to new vectors, or rather, paths while we are building them. Each
time a path is found a new one is allocated, built, and stored in the route_set.

p, connector: Scratch variables for iteration.

Optimisations: As described in the paper, both optimisations have been implemented. The path finding operates on the
compressed graph and the queue is shuffled if its possible to fill the route set that iteration. The route set may not
be filled due to duplicate paths but we can't know that in advance so we shuffle anyway.

Any further optimisations should focus on the path finding, from benchmarks it dominates the run time (~98%). Since huge
routes aren't required small-ish things like the memcpy and banned link set copy aren't high priority.

"""

# It would really be nice if these were modules. The 'include' syntax is long deprecated and adds a lot to compilation
# times
include 'basic_path_finding.pyx'


@cython.embedsignature(True)
cdef class RouteChoiceSet:
    """
    Route choice implemented via breadth first search with link removal (BFS-LE) as described in Rieser-Schüssler,
    Balmer, and Axhausen, 'Route Choice Sets for Very High-Resolution Data'
    """

    def __init__(self, graph: Graph):
        """Python level init, may be called multiple times, for things that can't be done in __cinit__."""
        # self.heuristic = HEURISTIC_MAP[self.res._heuristic]
        self.cost_view = graph.compact_cost
        self.graph_fs_view = graph.compact_fs
        self.b_nodes_view = graph.compact_graph.b_node.values
        self.nodes_to_indices_view = graph.compact_nodes_to_indices

        # tmp = graph.lonlat_index.loc[graph.compact_all_nodes]
        # self.lat_view = tmp.lat.values
        # self.lon_view = tmp.lon.values
        self.a_star = False

        self.ids_graph_view = graph.compact_graph.id.values

        # We explicitly don't want the links that have been removed from the graph
        self.graph_compressed_id_view = graph.graph.__compressed_id__.values
        self.num_nodes = graph.compact_num_nodes
        self.num_links = graph.compact_num_links
        self.zones = graph.num_zones
        self.block_flows_through_centroids = graph.block_centroid_flows

        self.mapping_idx, self.mapping_data, _ = graph.create_compressed_link_network_mapping()

        self.results = None
        self.ll_results = None

    @cython.embedsignature(True)
    def run(self, origin: int, destination: int, shape: Tuple[int, int], demand: float = 0.0, *args, **kwargs):
        """Compute the a route set for a single OD pair.

        Often the returned list's length is ``max_routes``, however, it may be limited by ``max_depth`` or if all
        unique possible paths have been found then a smaller set will be returned.

        Additional arguments are forwarded to ``RouteChoiceSet.batched``.

        :Arguments:
            **origin** (:obj:`int`): Origin node ID. Must be present within compact graph. Recommended to choose a
                centroid.
            **destination** (:obj:`int`): Destination node ID. Must be present within compact graph. Recommended to
                choose a centroid.
            **demand** (:obj:`double`): Demand for this single OD pair.

        :Returns: **route set** (:obj:`list[tuple[int, ...]]): Returns a list of unique variable length tuples of
            link IDs. Represents paths from ``origin`` to ``destination``.
        """
        df = pd.DataFrame({
            "origin id": [origin],
            "destination id": [destination],
            "demand": [demand]
        }).set_index(["origin id", "destination id"])
        demand_coo = GeneralisedCOODemand("origin id", "destination id", np.asarray(self.nodes_to_indices_view), shape)
        demand_coo.add_df(df)

        self.batched(demand_coo, {}, *args, **kwargs)
        where = kwargs.get("where", None)
        if where is not None:
            results = RouteChoiceSetResults.read_dataset(where)
        else:
            results = self.get_results()
        return [tuple(x) for x in results["route set"]]

    # Bounds checking doesn't really need to be disabled here but the warning is annoying
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.initializedcheck(False)
    def batched(
        self,
        demand: GeneralisedCOODemand,
        select_links: Dict[str, FrozenSet[FrozenSet[int]]] = None,
        sl_link_loading: bool = True,
        max_routes: int = 0,
        max_depth: int = 0,
        max_misses: int = 100,
        seed: int = 0,
        cores: int = 0,
        a_star: bool = True,
        bfsle: bool = True,
        penalty: float = 1.0,
        where: Optional[str] = None,
        store_results: bool = True,
        path_size_logit: bool = False,
        beta: float = 1.0,
        cutoff_prob: float = 0.0,
    ):
        """Compute the a route set for a list of OD pairs.

        Often the returned list for each OD pair's length is ``max_routes``, however, it may be limited by ``max_depth``
        or if all unique possible paths have been found then a smaller set will be returned.

        :Arguments:
            **ods** (:obj:`list[tuple[int, int]]`): List of OD pairs ``(origin, destination)``. Origin and destination
                node ID must be present within compact graph. Recommended to choose a centroids.
            **max_routes** (:obj:`int`): Maximum size of the generated route set. Must be non-negative. Default of
                ``0`` for unlimited.
            **max_depth** (:obj:`int`): Maximum depth BFSLE can explore, or maximum number of iterations for link
                penalisation. Must be non-negative. Default of ``0`` for unlimited.
            **max_misses** (:obj:`int`): Maximum number of collective duplicate routes found for a single OD pair.
                Terminates if exceeded.
            **seed** (:obj:`int`): Seed used for rng. Must be non-negative. Default of ``0``.
            **cores** (:obj:`int`): Number of cores to use when parallelising over OD pairs. Must be non-negative.
                Default of ``0`` for all available.
            **bfsle** (:obj:`bool`): Whether to use Breadth First Search with Link Removal (BFSLE) over link
                penalisation. Default ``True``.
            **penalty** (:obj:`float`): Penalty to use for Link Penalisation and BFSLE with LP.
            **where** (:obj:`str`): Optional file path to save results to immediately. Will return None.
        """
        cdef:
            long long origin, dest
            long int i

        if select_links is None:
            select_links = {}

        if max_routes == 0 and max_depth == 0:
            raise ValueError("Either `max_routes` or `max_depth` must be > 0")

        if max_routes < 0 or max_depth < 0:
            raise ValueError("`max_routes`, `max_depth`, and `cores` must be non-negative")

        if path_size_logit and beta < 0:
            raise ValueError("`beta` must be >= 0 for path sized logit model")

        if path_size_logit and not 0.0 <= cutoff_prob <= 1.0:
            raise ValueError("`cutoff_prob` must be 0 <= `cutoff_prob` <= 1 for path sized logit model")

        for origin, dest in demand.df.index:
            if self.nodes_to_indices_view[origin] == -1:
                raise ValueError(f"Origin {origin} is not present within the compact graph")
            if self.nodes_to_indices_view[dest] == -1:
                raise ValueError(f"Destination {dest} is not present within the compact graph")

        cdef:
            long long origin_index, dest_index
            unsigned int c_max_routes = max_routes
            unsigned int c_max_depth = max_depth
            unsigned int c_max_misses = max_misses
            unsigned int c_seed = seed
            long int c_cores = cores if cores > 0 else omp_get_max_threads()

            # Scale cutoff prob from [0, 1] -> [0.5, 1]. Values below 0.5 produce negative inverse binary logit values.
            double scaled_cutoff_prob = (1.0 - cutoff_prob) * 0.5 + 0.5

            # A* (and Dijkstra's) require memory views, so we must allocate here and take slices. Python can handle this
            # memory
            double [:, :] cost_matrix = np.empty((c_cores, self.cost_view.shape[0]), dtype=float)
            long long [:, :] predecessors_matrix = np.empty((c_cores, self.num_nodes + 1), dtype=np.int64)
            long long [:, :] conn_matrix = np.empty((c_cores, self.num_nodes + 1), dtype=np.int64)
            long long [:, :] b_nodes_matrix = np.broadcast_to(
                self.b_nodes_view,
                (c_cores, self.b_nodes_view.shape[0])
            ).copy()

            # This matrix is never read from, it exists to allow using the Dijkstra's method without changing the
            # interface.
            long long [:, :] _reached_first_matrix

            unsigned char [:, :] destinations_matrix = np.zeros((c_cores, self.num_nodes), dtype="bool")

            # self.a_star = a_star

        if self.a_star:
            _reached_first_matrix = np.zeros((c_cores, 1), dtype=np.int64)  # Dummy array to allow slicing
        else:
            _reached_first_matrix = np.zeros((c_cores, self.num_nodes + 1), dtype=np.int64)

        cdef:
            RouteSet_t *route_set
            shared_ptr[vector[double]] prob_vec
            int thread_id
            bint found_zero_cost

        demand._initalise_col_names()
        self.ll_results = LinkLoadingResults(demand, select_links, self.num_links, sl_link_loading, c_cores)

        # These are accessed with the gil and used for error reporting
        zero_cost_ods: list[tuple[int]] = []
        unreachable_ods: list[tuple[int]] = []

        for _, grouped_demand_df in (demand.batches() if where is not None else ((None, None),)):
            demand._initalise_c_data(grouped_demand_df)

            self.results = RouteChoiceSetResults(
                demand,
                scaled_cutoff_prob,
                beta,
                self.num_links,
                self.cost_view,
                self.mapping_idx,
                self.mapping_data,
                store_results=store_results,
                perform_assignment=path_size_logit,
            )

            with nogil, parallel(num_threads=c_cores):
                route_set = new RouteSet_t()
                thread_id = threadid()
                found_zero_cost = False  # Make the variable thread local
                for i in prange(<long int>demand.ods.size(), schedule="guided"):
                    origin_index = self.nodes_to_indices_view[demand.ods[i].first]
                    dest_index = self.nodes_to_indices_view[demand.ods[i].second]

                    if origin_index == dest_index:
                        continue

                    if self.block_flows_through_centroids:
                        blocking_centroid_flows(
                            0,  # Always blocking
                            origin_index,
                            self.zones,
                            self.graph_fs_view,
                            b_nodes_matrix[thread_id],
                            self.b_nodes_view,
                        )

                    if bfsle:
                        RouteChoiceSet.bfsle(
                            self,
                            d(route_set),
                            origin_index,
                            dest_index,
                            c_max_routes,
                            c_max_depth,
                            c_max_misses,
                            cost_matrix[thread_id],
                            predecessors_matrix[thread_id],
                            conn_matrix[thread_id],
                            b_nodes_matrix[thread_id],
                            _reached_first_matrix[thread_id],
                            destinations_matrix[thread_id],
                            penalty,
                            c_seed,
                        )
                    else:
                        RouteChoiceSet.link_penalisation(
                            self,
                            d(route_set),
                            origin_index,
                            dest_index,
                            c_max_routes,
                            c_max_depth,
                            c_max_misses,
                            cost_matrix[thread_id],
                            predecessors_matrix[thread_id],
                            conn_matrix[thread_id],
                            b_nodes_matrix[thread_id],
                            _reached_first_matrix[thread_id],
                            destinations_matrix[thread_id],
                            penalty,
                            c_seed,
                        )

                    # Here we transform the set of raw pointers to routes (vectors) into a vector of unique points to
                    # routes. This is done to simplify memory management later on.
                    route_vec = self.results.get_route_vec(i)
                    RouteChoiceSetResults.route_set_to_route_vec(d(route_vec), d(route_set))

                    if path_size_logit:
                        prob_vec = self.results.compute_result(i, d(route_vec), &found_zero_cost, thread_id)
                        self.ll_results.link_load_single_route_set(i, d(route_vec), d(prob_vec), thread_id)
                        self.ll_results.sl_link_load_single_route_set(
                            i, d(route_vec),
                            d(prob_vec),
                            origin_index,
                            dest_index,
                            thread_id
                        )


                    if d(route_vec).size() == 0 or found_zero_cost:
                        with gil:
                            if found_zero_cost:
                                zero_cost_ods.append(tuple(demand.ods[i]))
                            if d(route_vec).size() == 0:
                                unreachable_ods.append(tuple(demand.ods[i]))

                    if self.block_flows_through_centroids:
                        blocking_centroid_flows(
                            1,  # Always unblocking
                            origin_index,
                            self.zones,
                            self.graph_fs_view,
                            b_nodes_matrix[thread_id],
                            self.b_nodes_view,
                        )

                del route_set

            if store_results:
                self.get_results()
                if where is not None:
                    self.results.write(where)

        if path_size_logit:
            self.ll_results.reduce_link_loading()
            self.ll_results.reduce_sl_link_loading()
            self.ll_results.reduce_sl_od_matrix()

            self.get_link_loading(cores=c_cores)
            self.get_sl_link_loading(cores=c_cores)
            self.get_sl_od_matrices()

        if zero_cost_ods:
            warnings.warn(
                f"found zero cost routes for OD pairs, the entire route set has been masked for: {zero_cost_ods}"
            )
        if unreachable_ods:
            warnings.warn(
                f"found unreachable OD pairs, no choice sets generated for: {unreachable_ods}"
            )

    def assign_from_df(
        self,
        graph: pd.DataFrame,
        df: pd.DataFrame,
        demand: GeneralisedCOODemand,
        select_links: Dict[str, FrozenSet[FrozenSet[int]]] = None,
        recompute_psl: bool = False,
        sl_link_loading: bool = True,
        store_results: bool = True,
        beta: float = 1.0,
        cutoff_prob: float = 0.0,
    ):
        cdef:
            long int c_cores = 1  # Single threaded only due to high python interop, this shoud be fast anyway
            int thread_id = 0

            # Scale cutoff prob from [0, 1] -> [0.5, 1]. Values below 0.5 produce negative inverse binary logit values.
            double scaled_cutoff_prob = (1.0 - cutoff_prob) * 0.5 + 0.5

        for _, route_list in df["route set"].items():
            if not isinstance(route_list, (list, np.ndarray)):
                raise TypeError(f"route sets must be a list or Numpy array, found {type(route_list)}")

        # We want to enforce that if the demand matrix cell is non-cell for an OD pair, then at least one route exists to assign to it
        demand_df = demand.df.assign(idx=np.arange(len(demand.df)))
        demand_df = demand_df[demand_df.index.get_level_values(0) != demand_df.index.get_level_values(1)]

        df = df.set_index(demand_df.index.names)
        if not demand_df.index.drop_duplicates().isin(df.index).all():
            raise KeyError("not all origin and destinations IDs from the demand matrix are present within the path files")

        # We also store those indices along side the route sets themselves so it's easier to keep track
        df = demand_df[["idx"]].merge(df, how="left", left_index=True, right_index=True).reset_index()
        gb = df.groupby(by="idx")

        # In order to map the network link IDs to compressed links we'll use the graph
        graph_to_compressed = graph[["link_id", "direction"]].prod(axis=1).reset_index().set_index(0)

        # Now we initialise the demand matrix and prepare to insert the route sets
        demand._initalise_col_names()
        demand._initalise_c_data(None)

        self.results = RouteChoiceSetResults(
            demand,
            scaled_cutoff_prob,
            beta,
            self.num_links,
            self.cost_view,
            self.mapping_idx,
            self.mapping_data,
            store_results=store_results,
            perform_assignment=True,
        )

        self.ll_results = LinkLoadingResults(demand, select_links, self.num_links, sl_link_loading, c_cores)

        cdef:
            vector[long long] *route
            bint found_zero_cost

        # We iterate over the OD pairs in the path files
        for od_idx, df in gb:
            # We obtain a reference to the route vector, we then need to insert the right *compressed* link IDs
            route_vec = self.results.get_route_vec(od_idx)

            # If we are reusing the probabilities then we need to a similar thing for this
            if not recompute_psl:
                prob_vec = self.results.get_prob_vec(od_idx)

            d(route_vec).reserve(len(df))
            for _, row in df.iterrows():
                # We find the indices for the compressed id that corresponds to the direction link id pair (as a
                # product)
                compressed_link_indices = graph_to_compressed.loc[row["route set"]]["index"].to_numpy()

                route = new vector[long long]()
                # Then use itertools.groupby to de-duplicate them without modifying the order. The order is not required
                # for assignment but it is if we wish to output this route set again.
                for compressed_link_id, _ in itertools.groupby(graph.__compressed_id__.iloc[compressed_link_indices]):
                    route.push_back(compressed_link_id)

                d(route_vec).emplace_back(route)
                if not recompute_psl:
                    d(prob_vec).push_back(row["probability"])

            # If we are recomputing the probabilities then we do so here. This also has the side effect of recompute the
            # cost, masking, and path overlap with new parameters
            if recompute_psl:
                prob_vec = self.results.compute_result(od_idx, d(route_vec), &found_zero_cost, thread_id)

            # We have now have both the route and probability vectors restored so we can do LL and SLL.
            self.ll_results.link_load_single_route_set(od_idx, d(route_vec), d(prob_vec), thread_id)

            origin_index = self.nodes_to_indices_view[demand.ods[od_idx].first]
            dest_index = self.nodes_to_indices_view[demand.ods[od_idx].second]
            self.ll_results.sl_link_load_single_route_set(
                od_idx,
                d(route_vec),
                d(prob_vec),
                origin_index,
                dest_index,
                thread_id
            )

        # Clean up and reduce any results from the threaded storage
        self.ll_results.reduce_link_loading()
        self.ll_results.reduce_sl_link_loading()
        self.ll_results.reduce_sl_od_matrix()

        self.get_link_loading(cores=c_cores)
        self.get_sl_link_loading(cores=c_cores)
        self.get_sl_od_matrices()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.initializedcheck(False)
    cdef void path_find(
        RouteChoiceSet self,
        long origin_index,
        long dest_index,
        double [:] thread_cost,
        long long [:] thread_predecessors,
        long long [:] thread_conn,
        long long [:] thread_b_nodes,
        long long [:] _thread_reached_first,
        unsigned char [:] thread_destinations
    ) noexcept nogil:
        """Small wrapper around path finding, thread locals should be passes as arguments."""
        if self.a_star:
            path_finding_a_star(
                origin_index,
                dest_index,
                thread_cost,
                thread_b_nodes,
                self.graph_fs_view,
                self.nodes_to_indices_view,
                self.lat_view,
                self.lon_view,
                thread_predecessors,
                self.ids_graph_view,
                thread_conn,
                EQUIRECTANGULAR  # FIXME: enum import failing due to redefinition
            )
        else:
            thread_destinations[dest_index] = True
            path_finding(
                origin_index,
                thread_destinations,
                1,  # Single destination
                thread_cost,
                thread_b_nodes,
                self.graph_fs_view,
                thread_predecessors,
                self.ids_graph_view,
                thread_conn,
                _thread_reached_first
            )
            thread_destinations[dest_index] = False

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.initializedcheck(False)
    cdef void bfsle(
        RouteChoiceSet self,
        RouteSet_t &route_set,
        long origin_index,
        long dest_index,
        unsigned int max_routes,
        unsigned int max_depth,
        unsigned int max_misses,
        double [:] thread_cost,
        long long [:] thread_predecessors,
        long long [:] thread_conn,
        long long [:] thread_b_nodes,
        long long [:] _thread_reached_first,
        unsigned char [:] thread_destinations,
        double penalty,
        unsigned int seed
    ) noexcept nogil:
        """Main method for route set generation. See top of file for commentary."""
        cdef:
            # Scratch objects
            LinkSet_t removed_links
            minstd_rand rng

            # These objects are juggled to prevent more allocations than necessary
            vector[unordered_set[long long] *] queue
            vector[unordered_set[long long] *] next_queue
            unordered_set[long long] *banned
            unordered_set[long long] *new_banned

            # Local variables, Cython doesn't allow conditional declarations
            vector[long long] *vec
            pair[RouteSet_t.iterator, bool] status
            pair[LinkSet_t.iterator, bool] banned_status
            unsigned int miss_count = 0
            long long p, connector

            # Link penalisation, only used when penalty != 1.0
            bint lp = penalty != 1.0
            vector[double] *penalised_cost = <vector[double] *>nullptr
            vector[double] *next_penalised_cost = <vector[double] *>nullptr

            # Because we can have duplicate banned link sets, the insertion may fail, in that case we free the set
            # immediately. However, by doing so we then can't tell (without a method to track it), which sets have
            # already been freed in the queue if we happened to early exit from it, so we use another variable to just
            # free the remaining items in the queue.
            bool free_remaining = False

        max_routes = max_routes if max_routes != 0 else UINT_MAX
        max_depth = max_depth if max_depth != 0 else UINT_MAX

        queue.push_back(new unordered_set[long long]())  # Start with no edges banned
        rng.seed(seed)

        if lp:
            # Although we don't need the dynamic ability of vectors here, Cython doesn't have the std::array module.
            penalised_cost = new vector[double](self.cost_view.shape[0])
            next_penalised_cost = new vector[double](self.cost_view.shape[0])
            copy(&self.cost_view[0], &self.cost_view[0] + self.cost_view.shape[0], penalised_cost.begin())
            copy(&self.cost_view[0], &self.cost_view[0] + self.cost_view.shape[0], next_penalised_cost.begin())

        # We'll go at most `max_depth` iterations down, at each depth we maintain a queue of the next set of banned
        # edges to consider
        for depth in range(max_depth):
            if miss_count > max_misses or route_set.size() >= max_routes or queue.size() == 0:
                break

            # If we could potentially fill the route_set after this depth, shuffle the queue
            if queue.size() + route_set.size() >= max_routes:
                shuffle(queue.begin(), queue.end(), rng)

            next_queue.clear()
            for banned in queue:
                if free_remaining:
                    del banned
                    continue

                if lp:
                    # We copy the penalised cost buffer into the thread cost buffer to allow us to apply link
                    # penalisation,
                    copy(penalised_cost.cbegin(), penalised_cost.cend(), &thread_cost[0])
                else:
                    # ...otherwise we just copy directly from the cost view.
                    memcpy(&thread_cost[0], &self.cost_view[0], self.cost_view.shape[0] * sizeof(double))

                for connector in d(banned):
                    thread_cost[connector] = INFINITY

                RouteChoiceSet.path_find(
                    self,
                    origin_index,
                    dest_index,
                    thread_cost,
                    thread_predecessors,
                    thread_conn,
                    thread_b_nodes,
                    _thread_reached_first,
                    thread_destinations
                )

                # Mark this set of banned links as seen
                banned_status = removed_links.insert(banned)
                if not banned_status.second:
                    # If we failed to insert this banned set then an equal set already exists within the removed links
                    del banned
                    banned = d(banned_status.first)

                # If the destination is reachable we must build the path and readd
                if thread_predecessors[dest_index] >= 0:
                    vec = new vector[long long]()
                    # Walk the predecessors tree to find our path, we build it up in a C++ vector because we can't know
                    # how long it'll be
                    p = dest_index
                    while p != origin_index:
                        connector = thread_conn[p]
                        p = thread_predecessors[p]
                        vec.push_back(connector)

                    if lp:
                        # Here we penalise all seen links for the *next* depth. If we penalised on the current depth
                        # then we would introduce a bias for earlier seen paths
                        for connector in d(vec):
                            # *= does not work
                            d(next_penalised_cost)[connector] = penalty * d(next_penalised_cost)[connector]

                    reverse(vec.begin(), vec.end())

                    for connector in d(vec):
                        # This is one area for potential improvement. Here we construct a new set from the old one,
                        # copying all the elements then add a single element. An incremental set hash function could be
                        # of use. However, the since of this set is directly dependent on the current depth and as the
                        # route set size grows so incredibly fast the depth will rarely get high enough for this to
                        # matter. Copy the previously banned links, then for each vector in the path we add one and
                        # push it onto our queue
                        new_banned = new unordered_set[long long](d(banned))
                        new_banned.insert(connector)
                        # If we've already seen this set of removed links before we already know what the path is and
                        # its in our route set.
                        if removed_links.find(new_banned) != removed_links.end():
                            del new_banned
                        else:
                            next_queue.push_back(new_banned)

                    # The de-duplication of routes occurs here
                    status = route_set.insert(vec)
                    if not status.second:
                        del vec  # If the insertion failed, free this vector, we already have one that is equal to it
                        miss_count = miss_count + 1

                    if miss_count > max_misses or route_set.size() >= max_routes:
                        free_remaining = True
                        continue  # This condition will be hit again at the start of the loop, we just don't want to
                                  # iterate over the rest of the things in queue when we know there is not more space.
                else:
                    pass

            queue.swap(next_queue)

            if lp:
                # Update the penalised_cost vector, since next_penalised_cost is always the one updated we just need to
                # bring penalised_cost up to date.
                copy(next_penalised_cost.cbegin(), next_penalised_cost.cend(), penalised_cost.begin())

        # We may have added more banned link sets to the queue then found out we hit the max depth, we should free those
        for banned in queue:
            del banned

        # We should also free all the sets in next_queue, we don't be needing them.  We remove next_queue before
        # removed_links because we just swapped it with queue, and removed_links contains a subset of those that were
        # added to queue (pre-swap). It may share elements so we make sure to erase them from the set before freeing
        # them to avoid a use-after free.

        for banned in removed_links:
            del banned

        if lp:
            # If we had enabled link penalisation, we'll need to free those vectors as well
            del penalised_cost
            del next_penalised_cost

    @cython.wraparound(False)
    @cython.embedsignature(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void link_penalisation(
        RouteChoiceSet self,
        RouteSet_t &route_set,
        long origin_index,
        long dest_index,
        unsigned int max_routes,
        unsigned int max_depth,
        unsigned int max_misses,
        double [:] thread_cost,
        long long [:] thread_predecessors,
        long long [:] thread_conn,
        long long [:] thread_b_nodes,
        long long [:] _thread_reached_first,
        unsigned char [:] thread_destinations,
        double penalty,
        unsigned int seed
    ) noexcept nogil:
        """Link penalisation algorithm for choice set generation."""
        cdef:
            # Scratch objects
            vector[long long] *vec
            long long p, connector
            pair[RouteSet_t.iterator, bool] status
            unsigned int miss_count = 0

        max_routes = max_routes if max_routes != 0 else UINT_MAX
        max_depth = max_depth if max_depth != 0 else UINT_MAX
        memcpy(&thread_cost[0], &self.cost_view[0], self.cost_view.shape[0] * sizeof(double))

        for depth in range(max_depth):
            if route_set.size() >= max_routes:
                break

            RouteChoiceSet.path_find(
                self,
                origin_index,
                dest_index,
                thread_cost,
                thread_predecessors,
                thread_conn,
                thread_b_nodes,
                _thread_reached_first,
                thread_destinations
            )

            if thread_predecessors[dest_index] >= 0:
                vec = new vector[long long]()
                # Walk the predecessors tree to find our path, we build it up in a C++ vector because we can't know how
                # long it'll be
                p = dest_index
                while p != origin_index:
                    connector = thread_conn[p]
                    p = thread_predecessors[p]
                    vec.push_back(connector)

                for connector in d(vec):
                    thread_cost[connector] = penalty * thread_cost[connector]

                reverse(vec.begin(), vec.end())

                # To prevent runaway algorithms if we find N duplicate routes we should stop
                status = route_set.insert(vec)
                if not status.second:
                    del vec  # If the insertion failed, free this vector, we already have one that is equal to it
                    miss_count = miss_count + 1

                if miss_count > max_misses:
                    break
            else:
                break

    def get_results(self):
        """
        :Returns:
            **route sets** (:obj:`pa.DataFrame`): Returns a table of OD pairs to lists of link IDs for
                each OD pair provided (as columns). Represents paths from ``origin`` to ``destination``.
        """
        if self.results is None:
            raise RuntimeError("Route Choice results not computed yet")

        return self.results.make_df_from_results()

    def get_link_loading(RouteChoiceSet self, cores: int = 0):
        """
        :Returns:
            **link loading results** (:obj:`Dict[str, np.array]`): Returns a dict of demand column names to
                uncompressed link loads.
        """
        if self.ll_results is None:
            raise RuntimeError("Link loading results not computed yet")

        return self.ll_results.link_loading_to_objects(
            self.graph_compressed_id_view,
            cores if cores > 0 else omp_get_max_threads()
        )

    def get_sl_link_loading(RouteChoiceSet self, cores: int = 0):
        """
        :Returns:
            **select link loading results** (:obj:`Dict[str, Dict[str, np.array]]`): Returns a dict of select link set
                names to a dict of demand column names to uncompressed select link loads.
        """
        if self.ll_results is None:
            raise RuntimeError("Link loading results not computed yet")

        return self.ll_results.sl_link_loading_to_objects(
            self.graph_compressed_id_view,
            cores if cores > 0 else omp_get_max_threads()
        )

    def get_sl_od_matrices(RouteChoiceSet self):
        """
        :Returns:
            **select link OD matrix results** (:obj:`Dict[str, Dict[str, scipy.sparse.coo_matrix]]`): Returns a dict of
                select link set names to a dict of demand column names to a sparse OD matrix
        """
        if self.ll_results is None:
            raise RuntimeError("Link loading results not computed yet")

        return self.ll_results.sl_od_matrices_structs_to_objects()

    def write_path_files(RouteChoiceSet self, where):
        """
        Write the path-files to the directory specified

        :Arguments:
            **where** (:obj:`pathlib.Path`): Directory to save the dataset to.
        """
        if self.results is None:
            raise RuntimeError("Route Choice results not computed yet")

        self.results.write(where)



from sqlite3 import Connection
from shapely.geometry import MultiLineString
from aequilibrae.transit.constants import Constants, ROUTE_ID_MULTIPLIER, AGENCY_MULTIPLIER
from aequilibrae.transit.transit_elements.basic_element import BasicPTElement


class Route(BasicPTElement):
    """Transit route element to feed into Transit_routes

    * route_id (:obj:`str`): ID of this route, starting with the agency prefix ID
    * route_short_name (:obj:`str`): Short name as found in the GTFS feed
    * route_long_name (:obj:`str`): Long name as found in the GTFS feed
    * route_desc (:obj:`str`): Route description as found in the GTFS feed
    * route_type (:obj:`int`):  Route type (mode) for this transit link
    * route_url (:obj:`str`): Route URL as found in the GTFS feed
    * route_color (:obj:`str`): Route color for mapping as found in the GTFS feed
    * route_text_color (:obj:`str`): Route color (text) for mapping as found in the GTFS feed
    * route_sort_order (:obj:`int`): Route rendering order as found in the GTFS feed
    * agency_id (:obj:`str`): Agency ID
    * pce (:obj:`float`): Vehicle PCE for this route
    * seated_capacity (:obj:`float`): Vehicle seated capacity for this route
    * total_capacity (:obj:`float`): Total vehicle capacity for this route"""

    def __init__(self, agency_id):
        self.route_id = -1
        self.route = ""
        self.route_short_name = ""
        self.route_long_name = ""
        self.route_desc = ""
        self.route_type = 0
        self.route_url = ""
        self.route_color = ""
        self.route_text_color = ""
        self.route_sort_order = 0
        self.agency_id = agency_id

        # Not part of GTFS
        self.pattern_id = 0
        self.pattern = ""
        self.pce = 2.0
        self.seated_capacity = 0
        self.total_capacity = 0
        self.shape: MultiLineString
        self.srid = -1
        self.number_of_cars = 0
        self.__sql = """insert into routes (route_id, pattern_id, route, agency_id, shortname, longname, description,
                                            route_type, pce, seated_capacity, total_capacity{})
                        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?{});"""
        self.sql = self.__sql
        self.__get_route_id()

    def populate(self, record: tuple, headers: list) -> None:
        for key, value in zip(headers, record):
            if key not in self.__dict__.keys():
                raise KeyError(f"{key} field in Routes.txt is unknown field for that file on GTFS")

            # We convert route_id into route, as the the GTFS header for it is not maintained in our database
            key = "route" if key == "route_id" else key
            self.__dict__[key] = value

    def save_to_database(self, conn: Connection, commit=True) -> None:
        """Saves route to the database"""

        data = self.data
        conn.execute(self.sql, data)
        if commit:
            conn.commit()

    @property
    def data(self):
        data = [
            self.route_id,
            self.pattern_id,
            self.route,
            self.agency_id,
            self.route_short_name,
            self.route_long_name,
            self.route_desc,
            int(self.route_type),
            self.pce,
            self.seated_capacity,
            self.total_capacity,
        ]
        if self.shape is None:
            shape = ""
            geo_fld = ""
        else:
            geo_fld = ", geometry "
            shape = ", ST_Multi(GeomFromWKB(?, ?))"
            data.extend([self.shape.wkb, self.srid])
        self.sql = self.__sql.format(geo_fld, shape)
        return data

    def __get_route_id(self):
        c = Constants()
        val = c.routes.get(self.agency_id, AGENCY_MULTIPLIER * self.agency_id)
        c.routes[self.agency_id] = val + ROUTE_ID_MULTIPLIER
        self.route_id = c.routes[self.agency_id]



def list_tables_in_db(conn):
    """
    Return a list with all tables within a database.

    :Arguments:
         **conn** (:obj: `sqlite3.Connection`): database connection
    """
    return [x[0] for x in conn.execute("SELECT name FROM sqlite_master WHERE type ='table'").fetchall()]



import gc
import string
from pathlib import Path
from typing import List, Tuple

import geopandas as gpd
import numpy as np
import pandas as pd
from pandas import json_normalize
from shapely.geometry import Polygon

from aequilibrae.context import get_active_project
from aequilibrae.parameters import Parameters
from aequilibrae.project.project_creation import remove_triggers, add_triggers
from aequilibrae.utils.db_utils import commit_and_close, list_columns
from aequilibrae.utils.aeq_signal import SIGNAL, simple_progress
from aequilibrae.utils.interface.worker_thread import WorkerThread
from .model_area_gridding import geometry_grid


class OSMBuilder(WorkerThread):
    signal = SIGNAL(object)

    def __init__(self, data, project, model_area: Polygon, clean: bool) -> None:
        WorkerThread.__init__(self, None)

        project.logger.info("Preparing OSM builder")
        self.signal.emit(["set_text", "Preparing OSM builder"])

        self.project = project or get_active_project()
        self.logger = self.project.logger
        self.model_area = geometry_grid(model_area, 4326)
        self.path = self.project.path_to_file
        self.node_start = 10000
        self.clean = clean
        self.report = []
        self.__all_ltp = pd.DataFrame([])
        self.__link_id = 1
        self.__valid_links = []

        # Building shapely geometries makes the code surprisingly slower.
        self.node_df = data["nodes"]
        self.node_df.loc[:, "node_id"] = np.arange(self.node_start, self.node_start + self.node_df.shape[0])
        gc.collect()
        self.links_df = data["links"]

    def doWork(self):
        with commit_and_close(self.path, spatial=True) as conn:
            self.__update_table_structure(conn)
            self.importing_network(conn)

            self.logger.info("Cleaning things up")
            conn.execute(
                "DELETE FROM nodes WHERE node_id NOT IN (SELECT a_node FROM links union all SELECT b_node FROM links)"
            )
            conn.commit()
            self.__do_clean(conn)

        self.signal.emit(["finished"])

    def importing_network(self, conn):
        self.logger.info("Importing the network")
        node_count = pd.DataFrame(self.links_df["nodes"].explode("nodes")).assign(counter=1).groupby("nodes").count()

        self.node_df.osm_id = self.node_df.osm_id.astype(np.int64)
        self.node_df.set_index(["osm_id"], inplace=True)

        self.__process_link_chunk()

        self.logger.info("Geo-procesing links")
        geometries = []
        self.links_df.set_index(["osm_id"], inplace=True)

        for idx, link in simple_progress(self.links_df.iterrows(), self.signal, "Adding network links"):
            # How can I link have less than two points?
            if not isinstance(link["nodes"], list):
                self.logger.debug(f"OSM link/feature {idx} does not have a list of nodes.")
                continue

            if len(link["nodes"]) < 2:
                self.logger.debug(f"Link {idx} has less than two nodes. {link.nodes}")
                continue

            # The link is a straight line between two points
            # Or all midpoints are only part of a single link
            node_indices = node_count.loc[link["nodes"], "counter"].to_numpy()
            if len(link["nodes"]) == 2 or node_indices[1:-1].max() == 1:
                # The link has no intersections
                geometries.append([idx, self._build_geometry(link.nodes)])
            else:
                # Make sure we get the first and last nodes, as they are certainly the extremities of the sublinks
                node_indices[0] = 2
                node_indices[-1] = 2
                # The link has intersections
                # We build repeated records for links when they have intersections
                # This is because it is faster to do this way and then have all the data repeated
                # when doing the join with the link fields below
                intersecs = np.where(node_indices > 1)[0]
                for i, j in zip(intersecs[:-1], intersecs[1:]):
                    geometries.append([idx, self._build_geometry(link.nodes[i : j + 1])])

        # Builds the link Geo dataframe
        self.links_df.drop(columns=["nodes"], inplace=True)
        # We build a dataframe with the geometries created above
        # and join with the database
        geo_df = pd.DataFrame(geometries, columns=["link_id", "geometry"]).set_index("link_id")
        self.links_df = self.links_df.join(geo_df, how="inner")

        self.links_df.loc[:, "link_id"] = np.arange(self.links_df.shape[0]) + 1

        self.node_df = self.node_df.reset_index()

        # Saves the data to disk in case of issues loading it to the database
        osm_data_path = Path(self.project.project_base_path) / "osm_data"
        osm_data_path.mkdir(exist_ok=True)
        self.links_df.to_parquet(osm_data_path / "links.parquet")
        self.node_df.to_parquet(osm_data_path / "nodes.parquet")

        self.logger.info("Adding nodes to file")
        self.signal.emit(["set_text", "Adding nodes to file"])

        # Removing the triggers before adding all nodes makes things a LOT faster
        remove_triggers(conn, self.logger, "network")

        cols = ["node_id", "osm_id", "is_centroid", "modes", "link_types", "lon", "lat"]
        insert_qry = f"INSERT INTO nodes ({','.join(cols[:-2])}, geometry) VALUES(?,?,?,?,?, MakePoint(?,?, 4326))"
        conn.executemany(insert_qry, self.node_df[cols].to_records(index=False))

        del self.node_df
        gc.collect()

        # But we need to add them back to add the links
        add_triggers(conn, self.logger, "network")

        # self.links_df.to_file(self.project.path_to_file, driver="SQLite", spatialite=True, layer="links", mode="a")

        # I could not get the above line to work, so I used the following code instead
        self.links_df.index.name = "osm_id"
        self.links_df.reset_index(inplace=True)
        insert_qry = "INSERT INTO links ({},a_node, b_node, distance, geometry) VALUES({},0,0,0, GeomFromText(?, 4326))"
        cols_no_geo = self.links_df.columns.tolist()
        cols_no_geo.remove("geometry")
        insert_qry = insert_qry.format(", ".join(cols_no_geo), ", ".join(["?"] * len(cols_no_geo)))

        cols = cols_no_geo + ["geometry"]
        links_df = self.links_df[cols].to_records(index=False)

        del self.links_df
        gc.collect()
        self.logger.info("Adding links to file")
        self.signal.emit(["set_text", "Adding links to file"])
        conn.executemany(insert_qry, links_df)

    def _build_geometry(self, nodes: List[int]) -> str:
        slice = self.node_df.loc[nodes, :]
        txt = ",".join((slice.lon.astype(str) + " " + slice.lat.astype(str)).tolist())
        return f"LINESTRING({txt})"

    def __do_clean(self, conn):
        if not self.clean:
            conn.execute("VACUUM;")
            return
        self.logger.info("Cleaning up the network down to the selected area")
        links = gpd.GeoDataFrame.from_postgis("SELECT link_id, asBinary(geometry) AS geom FROM links", conn, crs=4326)
        existing_link_ids = gpd.sjoin(links, self.model_area, how="left").dropna().link_id.to_numpy()
        to_delete = [[x] for x in links[~links.link_id.isin(existing_link_ids)].link_id]
        conn.executemany("DELETE FROM links WHERE link_id = ?", to_delete)
        conn.commit()
        conn.execute("VACUUM;")

    def __process_link_chunk(self):
        self.logger.info("Processing link modes, types and fields")
        self.signal.emit(["set_text", "Processing link modes, types and fields"])

        # It is hard to define an optimal chunk_size, so let's assume that 1GB is a good size per chunk
        # And let's also assume that each row will be 200 fields at 8 bytes each
        # This makes 2Gb roughly equal to 2.6 million rows, so 2 million would so.
        chunk_size = 1_000_000
        list_dfs = [self.links_df.iloc[i : i + chunk_size] for i in range(0, self.links_df.shape[0], chunk_size)]
        self.links_df = []
        # Initialize link types
        with self.project.db_connection as conn:
            self.__all_ltp = pd.read_sql('SELECT link_type_id, link_type, "" as highway from link_types', conn)
            for df in simple_progress(list_dfs, self.signal, "Processing chunks"):
                if "tags" in df.columns:
                    # It is critical to reset the index for the concat below to work
                    df.reset_index(drop=True, inplace=True)
                    df = pd.concat([df, json_normalize(df["tags"])], axis=1).drop(columns=["tags"])
                    df.columns = [x.replace(":", "_") for x in df.columns]
                    df = self.__build_link_types(df)
                    df = self.__establish_modes_for_all_links(conn, df)
                    df = self.__process_link_attributes(df)
                else:
                    self.logger.error("OSM link data does not have tags. Skipping an entire data chunk")
                    df = pd.DataFrame([])
                self.links_df.append(df)
        self.links_df = pd.concat(self.links_df, ignore_index=True)

    def __build_link_types(self, df):
        data = []
        df = df.fillna(value={"highway": "missing"})
        df.highway = df.highway.str.lower()
        for lt in df.highway.unique():
            if str(lt) in self.__all_ltp.highway.values:
                continue
            data.append([*self.__define_link_type(str(lt)), str(lt)])
            self.__all_ltp = pd.concat(
                [self.__all_ltp, pd.DataFrame(data, columns=["link_type_id", "link_type", "highway"])]
            )
            self.__all_ltp.drop_duplicates(inplace=True)
        df = df.merge(self.__all_ltp[["link_type", "highway"]], on="highway", how="left")
        return df.drop(columns=["highway"])

    def __define_link_type(self, link_type: str) -> Tuple[str, str]:
        proj_link_types = self.project.network.link_types
        original_link_type = link_type
        link_type = "".join([x for x in link_type if x in string.ascii_letters + "_"]).lower()

        split = link_type.split("_")
        for i, piece in enumerate(split[1:]):
            if piece in ["link", "segment", "stretch"]:
                link_type = "_".join(split[0 : i + 1])

        if self.__all_ltp.shape[0] >= 51:
            link_type = "aggregate_link_type"

        if len(link_type) == 0:
            link_type = "empty"

        if link_type in self.__all_ltp.link_type.values:
            lt = proj_link_types.get_by_name(link_type)
            if original_link_type not in lt.description:
                lt.description += f", {original_link_type}"
                lt.save()
            return [lt.link_type_id, link_type]

        letter = link_type[0]
        if letter in self.__all_ltp.link_type_id.values:
            letter = letter.upper()
            if letter in self.__all_ltp.link_type_id.values:
                for letter in string.ascii_letters:
                    if letter not in self.__all_ltp.link_type_id.values:
                        break
        lt = proj_link_types.new(letter)
        lt.link_type = link_type
        lt.description = f"Link types from Open Street Maps: {original_link_type}"
        lt.save()
        return [letter, link_type]

    def __establish_modes_for_all_links(self, conn, df: pd.DataFrame) -> pd.DataFrame:
        p = Parameters()
        modes = p.parameters["network"]["osm"]["modes"]

        mode_codes = conn.execute("SELECT mode_name, mode_id from modes").fetchall()
        mode_codes = {p[0]: p[1] for p in mode_codes}

        type_list = {}
        notfound = ""
        for mode, val in modes.items():
            all_types = val["link_types"]
            md = mode_codes[mode]
            for tp in all_types:
                type_list[tp] = "".join(sorted("{}{}".format(type_list.get(tp, ""), md)))
            if val["unknown_tags"]:
                notfound += md

        type_list = {k: "".join(set(v)) for k, v in type_list.items()}

        df_aux = pd.DataFrame([[k, v] for k, v in type_list.items()], columns=["link_type", "modes"])
        df = df.merge(df_aux, on="link_type", how="left").fillna(value={"modes": "".join(sorted(notfound))})
        return df

    def __process_link_attributes(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(direction=0, link_id=0)
        if "oneway" in df.columns:
            df.loc[df.oneway == "yes", "direction"] = 1
            df.loc[df.oneway == "backward", "direction"] = -1
        p = Parameters()
        fields = p.parameters["network"]["links"]["fields"]

        for x in fields["one-way"]:
            if "link_type" in x.keys():
                continue
            keys_ = list(x.values())[0]
            field = list(x.keys())[0]
            osm_name = keys_.get("osm_source", field).replace(":", "_")
            df.rename(columns={osm_name: field}, inplace=True, errors="ignore")

        for x in fields["two-way"]:
            keys_ = list(x.values())[0]
            field = list(x.keys())[0]
            if "osm_source" not in keys_:
                continue
            osm_name = keys_.get("osm_source", field).replace(":", "_")
            if osm_name not in df.columns:
                continue
            df[f"{field}_ba"] = df[osm_name].copy()
            df.rename(columns={osm_name: f"{field}_ab"}, inplace=True, errors="ignore")
            if "osm_behaviour" in keys_ and keys_["osm_behaviour"] == "divide":
                df[f"{field}_ab"] = pd.to_numeric(df[f"{field}_ab"], errors="coerce")
                df[f"{field}_ba"] = pd.to_numeric(df[f"{field}_ba"], errors="coerce")

                # Divides the values by 2 or zero them depending on the link direction
                df.loc[df.direction == 0, f"{field}_ab"] /= 2
                df.loc[df.direction == -1, f"{field}_ab"] = 0

                df.loc[df.direction == 0, f"{field}_ba"] /= 2
                df.loc[df.direction == 1, f"{field}_ba"] = 0

                if f"{field}_forward" in df:
                    fld = pd.to_numeric(df[f"{field}_forward"], errors="coerce")
                    df.loc[fld > 0, f"{field}_ab"] = fld[fld > 0]
                if f"{field}_backward" in df:
                    fld = pd.to_numeric(df[f"{field}_backward"], errors="coerce")
                    df.loc[fld > 0, f"{field}_ba"] = fld[fld > 0]
        with self.project.db_connection as conn:
            cols = list_columns(conn, "links") + ["nodes"]
        return df[[x for x in cols if x in df.columns]]

    ######## TABLE STRUCTURE UPDATING ########
    def __update_table_structure(self, conn):
        structure = conn.execute("pragma table_info(Links)").fetchall()
        has_fields = [x[1].lower() for x in structure]
        fields = [field.lower() for field in self.get_link_fields()] + ["osm_id"]
        for field in [f for f in fields if f not in has_fields]:
            ltype = self.get_link_field_type(field).upper()
            conn.execute(f"Alter table Links add column {field} {ltype}")
        conn.commit()

    @staticmethod
    def get_link_fields():
        p = Parameters()
        fields = p.parameters["network"]["links"]["fields"]
        owf = [list(x.keys())[0] for x in fields["one-way"]]

        twf1 = ["{}_ab".format(list(x.keys())[0]) for x in fields["two-way"]]
        twf2 = ["{}_ba".format(list(x.keys())[0]) for x in fields["two-way"]]

        return owf + twf1 + twf2 + ["osm_id"]

    @staticmethod
    def get_link_field_type(field_name):
        p = Parameters()
        fields = p.parameters["network"]["links"]["fields"]

        if field_name[-3:].lower() in ["_ab", "_ba"]:
            field_name = field_name[:-3]
            for tp in fields["two-way"]:
                if field_name in tp:
                    return tp[field_name]["type"]
        else:
            for tp in fields["one-way"]:
                if field_name in tp:
                    return tp[field_name]["type"]



import logging
import os
import shutil
import urllib
import warnings
from os.path import join, basename
from pathlib import Path
from sqlite3 import Connection, register_adapter
from tempfile import gettempdir
from typing import Optional
from zipfile import ZipFile

import numpy as np

from aequilibrae.log import global_logger
from aequilibrae.utils.db_utils import has_table, safe_connect
from aequilibrae.utils.qgis_utils import inside_qgis

# Setup adapaters so that we can read/write numpy types directly to DB
register_adapter(np.int64, int)
register_adapter(np.int32, int)
register_adapter(np.float32, float)
register_adapter(np.float64, float)
register_adapter(np.object_, str)


def is_windows():
    return os.name == "nt"


def is_not_windows():
    return os.name != "nt"


def connect_spatialite(path_to_file: os.PathLike, missing_ok: bool = False) -> Connection:
    if inside_qgis:
        import qgis

        return qgis.utils.spatialite_connect(str(path_to_file))

    ensure_spatialite_binaries()

    return _connect_spatialite(path_to_file, missing_ok)


def _connect_spatialite(path_to_file: os.PathLike, missing_ok: bool = False):
    conn = safe_connect(path_to_file, missing_ok)
    load_spatialite_extension(conn)
    return conn


def load_spatialite_extension(conn: Connection):
    conn.enable_load_extension(True)
    conn.load_extension("mod_spatialite")


def is_spatialite(conn):
    return has_table(conn, "geometry_columns")


def set_known_spatialite_folder(spatialite_folder: os.PathLike):
    directory = str(spatialite_folder)
    if directory not in os.environ["PATH"]:
        os.environ["PATH"] = directory + os.pathsep + os.environ["PATH"]
    if "PROJ_LIB" not in os.environ:
        os.environ["PROJ_LIB"] = directory


def ensure_spatialite_binaries() -> None:
    if is_not_windows():
        return

    directory = os.environ.get("AEQ_SPATIALITE_DIR", gettempdir())

    if not _dll_already_exists(directory):
        global_logger.info(f"mod_spatialite.dll not found in {directory} attempting to download")
        _download_and_extract_spatialite(directory)

    set_known_spatialite_folder(directory)

    try:
        # We need to have the proj.db file in place.
        # The easiest one on Windows is in the public user. On Linux it should not be necessary
        # See why: https://www.gaia-gis.it/fossil/libspatialite/wiki?name=PROJ.6
        projdb_dir = "C:/Users/Public/spatialite/proj"
        Path(projdb_dir).mkdir(parents=True, exist_ok=True)
        if os.path.isfile(join(projdb_dir, "proj.db")):
            return

        shutil.copyfile(join(directory, "proj.db"), join(projdb_dir, "proj.db"))
    except Exception as e:
        msg = f"Could not put the proj.db file in the expected place. {e.args}"
        warnings.warn(msg)
        global_logger.warning(msg)


def _dll_already_exists(d: os.PathLike) -> bool:
    return os.path.exists(join(d, "mod_spatialite.dll"))


def _download_and_extract_spatialite(directory: os.PathLike) -> None:
    url = "https://github.com/AequilibraE/aequilibrae/releases/download/V.0.7.5/mod_spatialite-5.0.1-win-amd64.zip"
    zip_file = join(directory, basename(url))

    Path(directory).mkdir(exist_ok=True, parents=True)
    urllib.request.urlretrieve(url, zip_file)
    ZipFile(zip_file).extractall(directory)
    os.remove(zip_file)


def spatialize_db(conn, logger=None):
    logger = logger or logging.getLogger("aequilibrae")
    logger.info("Adding Spatialite infrastructure to the database")
    if not inside_qgis and not is_spatialite(conn):
        try:
            conn.execute("SELECT InitSpatialMetaData();")
            conn.commit()
        except Exception as e:
            logger.error("Problem with spatialite", e.args)
            raise e
    if not is_spatialite(conn):
        raise RuntimeError("Something went wrong")



from .agency_writer import write_agencies
from .routes_writer import write_routes
from .shape_writer import write_shapes
from .stop_times_writer import write_stop_times
from .stops_writer import write_stops
from .trips_writer import write_trips
from .fare_writer import write_fares



import numpy as np
cimport numpy as cnp

DTYPE = np.float64
ctypedef cnp.float64_t DTYPE_t

ITYPE = np.int64
ctypedef cnp.int64_t ITYPE_t

# EPS is the precision of DTYPE
cdef DTYPE_t DTYPE_EPS = 1E-15

# NULL_IDX is the index used in predecessor matrices to store a non-path
cdef ITYPE_t NULL_IDX = 18446744073709551615
